import logging
import os
from abc import ABC, abstractmethod
from glob import glob
from pathlib import Path

import numpy as np

from aspire.image import Image
from aspire.source import Simulation
from aspire.source.image import _ImageAccessor
from aspire.storage import StarFile
from aspire.utils import Random, grid_2d
from aspire.volume import Volume

logger = logging.getLogger(__name__)


class MicrographSource(ABC):
    def __init__(self, micrograph_count, micrograph_size, dtype):
        """ """
        self.micrograph_count = int(micrograph_count)
        self.micrograph_size = int(micrograph_size)
        self.dtype = np.dtype(dtype)

        self._images_accessor = _ImageAccessor(self._images, self.micrograph_count)

    def __repr__(self):
        """
        String representation.

        :return: Returns a string description of instance.
        """
        return f"{self.__class__.__name__} with {self.micrograph_count} {self.dtype.name} micrographs of size {self.micrograph_size}x{self.micrograph_size}"

    def __len__(self):
        """
        Returns the number of micrographs.

        :return: Returns the micrograph count
        """
        return self.micrograph_count

    def save(self, path, name_prefix="micrograph", overwrite=True):
        """
        Save micrographs to `path`.

        `path` should be a directory. If it does not exist, ASPIRE will attempt to create it.

        Currently saves micrographs to `.mrc`.

        :param path: Directory to save data.
        :param name_prefix: Optional, name prefix string for micrograph files.
        :param overwrite: Optional, bool. Allow writing to existing directory,
            and overwriting existing files.
        :return: List of saved `.mrc` files.
        """

        # Make dir if does not exist.
        Path(path).mkdir(parents=True, exist_ok=overwrite)

        output_mrc_filenames = []
        for m in range(self.micrograph_count):
            prefix = f"{name_prefix}_{m}"

            # Get the micrograph and save it as an `mrc`
            micrograph = self.images[m].asnumpy()

            mrc_file_path = os.path.join(path, f"{prefix}.mrc")
            Image(micrograph).save(mrc_file_path, overwrite=overwrite)

            output_mrc_filenames.append(mrc_file_path)

        return output_mrc_filenames

    def asnumpy(self):
        """
        Return micrograph data as dense Numpy array.

        :return: Numpy array.
        """
        return self.images[:]._data

    def show(self, *args, **kwargs):
        """
        Helper function to display micrograph. See Image.show().
        """
        Image(self.asnumpy()).show(*args, **kwargs)

    @property
    def images(self):
        """
        Returns micrographs.

        :return: An `ImageAccessor` for the noisy micrographs.
        """
        return self._images_accessor

    @abstractmethod
    def _images(self, indices):
        """
        Accesses and returns micrographs.

        :param indices: A 1-D Numpy array of integer indices.
        :return: An `Image` object representing the micrographs for `indices`.
        """


class ArrayMicrographSource(MicrographSource):
    def __init__(self, micrographs, dtype=None):
        """
        Instantiate a `MicrographSource` with `micrographs`.

        :param micrographs: Numpy array (count, L, L)
            where  `L` represents the size in pixels.
        :param dtype: Data type of returned `MicrographSource`.
            Default of `None` will infer dtype from the array.
            Explicitly setting dtype will attempt a cast.
            Currently only `float32` and `float64` are supported.
            Note, due to limitations of common MRC implementations,
            saving is limited to single precision.
        """

        # Check micrographs is an array
        if not isinstance(micrographs, np.ndarray):
            raise NotImplementedError(
                f"{self.__class__.__name__} not implemented for {type(micrographs)}."
            )

        # Ensure dimensions
        if micrographs.ndim == 2:
            micrographs = micrographs[None, :, :]

        if micrographs.ndim != 3 or (micrographs.shape[-2] != micrographs.shape[-1]):
            raise NotImplementedError(
                f"Incompatible `micrographs` shape {micrographs.shape}, expects (count, L, L)"
            )

        super().__init__(
            micrograph_count=micrographs.shape[0],
            micrograph_size=micrographs.shape[-1],
            dtype=dtype or micrographs.dtype,
        )

        # We're already backed by an array, access it directly.
        self._data = micrographs.astype(self.dtype, copy=False)

    def _images(self, indices):
        """
        Accesses and returns micrographs from a Numpy array.

        :param indices: A 1-D Numpy array of integer indices.
        :return: An array backed `MicrographSource` object representing the micrographs for `indices`.
        """
        return Image(self._data[indices])


class DiskMicrographSource(MicrographSource):
    def __init__(self, micrographs_path, dtype=None):
        """
        Instantiate a `MicrographSource` with `micrographs_path`.

        :param micrographs_path: Path string representing directory of file,
            or list of file path strings.
        :param dtype: Data type of returned `MicrographSource`.
            Default of `None` will attempt to infer dtype from the first micrograph.
            Explicitly setting dtype will attempt a cast when returning micrographs.
            Currently only `float32` and `float64` are supported.
            Note, due to limitations of common MRC implementations,
            saving is limited to single precision.
        """

        if not self._is_path_like_input(micrographs_path):
            raise NotImplementedError(
                f"{self.__class__.__name__} not implemented for {type(micrographs_path)}."
            )

        if hasattr(micrographs_path, "__len__") and len(micrographs_path) == 0:
            raise RuntimeError(
                f"Must supply non-empty `micrographs` argument, received {micrographs_path}."
            )

        if isinstance(micrographs_path, list):
            # Explicit list
            self.micrograph_files = micrographs_path
        else:
            # `str` cast will accommodate string and Path objects
            self.micrograph_files = self._glob_files(str(micrographs_path))

        # Load the first micrograph to infer shape/type
        # Size will be checked during on-the-fly loading of subsequent micrographs.
        micrograph0 = Image.load(self.micrograph_files[0])

        super().__init__(
            micrograph_count=len(self.micrograph_files),
            micrograph_size=micrograph0.resolution,
            dtype=dtype or micrograph0.dtype,
        )

        # Prepare accessor to load files from disk on the fly.
        self._images_accessor = _ImageAccessor(self._images, self.micrograph_count)

    @staticmethod
    def _is_path_like_input(p):
        """
        Utility to return whether `p` should be treated as a path to micrograph(s).

        :param p: Item to test.
        :return: Boolean.
        """
        return any([isinstance(p, list), isinstance(p, str), isinstance(p, Path)])

    def _glob_files(self, file_path):
        """
        Helper utility to glob for matching micrograph files at `file_path`.

        Will match on any image matching `Image.extensions` in `file_path`.

        :return: List of files.
        """

        files = []
        # Determine if directory or single file.
        if os.path.isdir(file_path):
            # Add all loadable images in the directory.
            for ext in Image.extensions:
                files.extend(sorted(glob(os.path.join(file_path, f"*{ext}"))))
            # Ensure we have not failed to find any micrographs in the provided directory.
            if len(files) == 0:
                raise RuntimeError(f"No suitable files were found at {file_path}.")
        elif os.path.isfile(file_path):
            # Add this file.
            files.append(file_path)
        else:
            raise RuntimeError(
                f"File path {file_path} does not appear to be a directory or a file"
            )

        return files

    def _images(self, indices):
        """
        Accesses and returns micrographs from disk.

        :param indices: A 1-D Numpy array of integer indices.
        :return: An `Image` object representing the micrographs for `indices`.
        """
        # Initialize empty result
        n_micrographs = len(indices)
        micrographs = np.empty(
            (n_micrographs, self.micrograph_size, self.micrograph_size),
            dtype=self.dtype,
        )
        for i, ind in enumerate(indices):
            # Load the micrograph image from file
            micrograph = Image.load(self.micrograph_files[ind])
            # Assert size
            if micrograph.resolution != self.micrograph_size:
                raise NotImplementedError(
                    f"Micrograph {ind} has inconsistent shape {micrograph.shape},"
                    f" expected {(self.micrograph_size, self.micrograph_size)}."
                )
            # Assign to array, implicitly performs casting to dtype
            micrographs[i] = micrograph.asnumpy()

        return Image(micrographs)


class MicrographSimulation(MicrographSource):
    def __init__(
        self,
        volume,
        micrograph_size=4096,
        micrograph_count=1,
        particles_per_micrograph=100,
        particle_amplitudes=None,
        projection_angles=None,
        seed=None,
        ctf_filters=None,
        noise_adder=None,
        boundary=None,
        interparticle_distance=None,
    ):
        """
        A cryo-EM MicrographSimulation object that supplies micrographs.

        `dtype` and `particle_box_size` are inferred from `volume`, where `dtype` is the data type of the micrographs and `particle_box_size` is the size of the particle images.

        :param volume: `Volume` instance to be used in `Simulation`.
             An `(L,L,L)` `Volume` will generate `(L,L)` particle images.
        :param micrograph_size: Size of micrograph in pixels, defaults to 4096.
        :param micrograph_count: Number of micrographs to generate (integer). Defaults to 1.
        :param particles_per_micrograph: The amount of particles generated for each micrograph. Defaults to 10.
        :param particle_amplitudes: Optional, amplitudes to pass to `Simulation`.
             Default `None` uses `Simulation` defaults.
             When provided must be array with size `particles_per_micrograph * micrograph_count`.
        :param projection_angles: Optional, projection rotation angles to pass to `Simulation`.
             Default `None` uses `Simulation` defaults.
             When provided must have shape `(particles_per_micrograph * micrograph_count, 3)`.

        :param seed: Random seed.
        :param noise_adder: Append instance of NoiseAdder to generation pipeline.
        :param ctf_filters: Optional list of `Filter` objects to apply to particles.
            This list should be 1, n_micrographs, or particles_per_micrograph * micrograph_count.
            These will apply filters to all particles, per-micrograph, or per-particle respectively.
            Default `None` will not apply any additional filters.
        :param boundary: Set boundaries for particle centers, positive values move the boundary inward from the edge of the micrograph. Defaults to half of the particle size (particle_box_size // 2).
        :param interparticle_distance: Set minimum distance between particle centers, in pixels. Defaults to particle_box_size.
        :return: A MicrographSimulation object.
        """
        if not isinstance(volume, Volume):
            raise TypeError("`volume` should be of type `Volume`.")
        self.volume = volume

        self.seed = seed

        super().__init__(
            micrograph_count=micrograph_count,
            micrograph_size=micrograph_size,
            dtype=self.volume.dtype,
        )

        self.particle_box_size = volume.resolution  # L
        self.particles_per_micrograph = particles_per_micrograph
        self.total_particle_count = (
            self.micrograph_count * self.particles_per_micrograph
        )
        self.dtype = volume.dtype

        self.noise_adder = noise_adder

        if self.particle_box_size > micrograph_size:
            raise ValueError(
                "The micrograph size must be larger or equal to the `particle_box_size`."
            )

        if particle_amplitudes is not None:
            if (
                not isinstance(particle_amplitudes, int)
                and len(particle_amplitudes) != self.total_particle_count
            ):
                raise RuntimeError(
                    f"`particle_amplitudes` must be an `int` or length {self.total_particle_count}."
                )
        self.particle_amplitudes = particle_amplitudes

        if projection_angles is not None and projection_angles.shape != (
            self.total_particle_count,
            3,
        ):
            raise RuntimeError(
                f"`projection_angles.shape` {projection_angles.shape} != (total_particle_count,3) {self.total_particle_count, 3}"
            )
        self.projection_angles = projection_angles

        self.filter_indices = None
        if ctf_filters is not None:
            acceptable_lens = [1, self.micrograph_count, self.total_particle_count]
            if (
                not isinstance(ctf_filters, list)
                or len(ctf_filters) not in acceptable_lens
            ):
                raise TypeError(
                    f"`ctf_filters` expects a list of len {acceptable_lens[0]},"
                    f" {acceptable_lens[1]}, or {acceptable_lens[2]}."
                )

            # Generate explicit filter indices (zero-indexed).
            self.filter_indices = np.arange(self.total_particle_count) % len(
                ctf_filters
            )

        self.ctf_filters = ctf_filters

        self.simulation = Simulation(
            n=self.total_particle_count,
            vols=self.volume,
            L=self.particle_box_size,
            offsets=0,
            amplitudes=self.particle_amplitudes,
            angles=self.projection_angles,
            unique_filters=ctf_filters,
            filter_indices=self.filter_indices,
            dtype=self.dtype,
            seed=self.seed,
        )

        if boundary is None:
            self.boundary = self.particle_box_size // 2
        else:
            if (
                boundary < (-self.particle_box_size // 2)
                or boundary > self.micrograph_size // 2
            ):
                raise ValueError("Illegal boundary value.")
            self.boundary = boundary

        if interparticle_distance is None:
            self.interparticle_distance = self.particle_box_size
        else:
            self.interparticle_distance = interparticle_distance

        # Create the radial mask for each center
        g2d = grid_2d(int(2 * self.interparticle_distance), normalized=False)
        radial_mask = g2d["r"] <= self.interparticle_distance
        self.grid_x = g2d["x"].astype(int)[radial_mask]
        self.grid_y = g2d["y"].astype(int)[radial_mask]

        # Calculate the proper padding for the micrograph borders and create the micrograph mask
        self.pad = int(max(self.particle_box_size, self.interparticle_distance))
        self._mask_boundary = self.pad + self.boundary
        self._set_mask()

        # Create the centers
        self.centers = np.zeros(
            (self.micrograph_count, self.particles_per_micrograph, 2), dtype=int
        )

        # Set pass threshold here
        self._fail_limit = 0.1 * self.micrograph_count
        self._fail_count = 0

        with Random(seed=self.seed) as _:
            for i in range(self.micrograph_count):
                self.centers[i] = self._create_centers(i)

        self._clean_images_accessor = _ImageAccessor(
            self._clean_images, self.micrograph_count
        )

        self._images_accessor = _ImageAccessor(self._images, self.micrograph_count)

    def _create_centers(self, micrograph_index):
        """
        Creates centers for the given micrograph if the fail threshold isn't met.

        param micrograph_index: The ID of the micrograph.
        return: A Numpy array containing the generated centers.
        """
        while self._fail_count < self._fail_limit:
            try:
                self._set_mask()
                centers = np.zeros((self.particles_per_micrograph, 2))
                for i in range(self.particles_per_micrograph):
                    x, y = self._generate_center()
                    centers[i] = np.array([x, y])
                return centers
            except RuntimeError:
                self._fail_count += 1
        else:
            raise RuntimeError(
                "Micrograph generation failures exceeded limit. This"
                "can happen if constraints are too strict. Consider"
                "adjusting micrograph_size, particles_per_micrograph,"
                " or interparticle_distance."
            )

    def _generate_center(self):
        """
        Helper method to generate centers using the mask.

        return: The x-y coordinate values of the generated center.
        """
        available_centers = np.transpose(np.where(self._mask))

        if available_centers.shape[0] == 0:
            self._fail_count += 1
            raise RuntimeError("Not enough centers generated.")

        random_index = np.random.choice(available_centers.shape[0])
        x, y = available_centers[random_index]
        x_vals = self.grid_x + x
        y_vals = self.grid_y + y
        self._mask[x_vals, y_vals] = False
        return x - self.pad, y - self.pad

    def _set_mask(self):
        """
        Helper method to set the mask.
        """
        self._mask = np.full(
            (
                int(self.micrograph_size + 2 * self.pad),
                int(self.micrograph_size + 2 * self.pad),
            ),
            False,
            dtype=bool,
        )
        self._mask[
            self._mask_boundary : -self._mask_boundary,
            self._mask_boundary : -self._mask_boundary,
        ] = True

    @property
    def clean_images(self):
        """
        Returns the micrographs without any noise.

        :return: An `ImageAccessor` for the unnoisy micrographs.
        """
        return self._clean_images_accessor

    def _images(self, indices):
        """
        Accesses and returns micrographs with any added noise.

        :param indices: A 1-D Numpy array of integer indices.
        :return: An `Image` object representing the noisy micrograph.
        """
        micrographs = self._clean_images(indices)
        if self.noise_adder:
            micrographs = self.noise_adder.forward(micrographs)
        return micrographs

    def _clean_images(self, indices):
        """
        Accesses and returns micrographs without any added noise.

        :param indices: A 1-D Numpy array of integer indices.
        :return: An `Image` object representing the clean micrograph
        """
        # Initialize empty micrograph
        n_micrographs = len(indices)
        clean_micrograph = np.zeros(
            (n_micrographs, self.micrograph_size, self.micrograph_size),
            dtype=self.dtype,
        )
        # Pad the micrograph
        clean_micrograph = np.pad(
            clean_micrograph,
            ((0, 0), (self.pad, self.pad), (self.pad, self.pad)),
            "constant",
            constant_values=(0),
        )
        # Get centers
        parity = self.particle_box_size % 2
        for m in range(n_micrographs):
            global_id = indices[m]
            images = self.simulation.clean_images[self.get_particle_indices(global_id)]
            centers = self.centers[global_id]
            x_lefts = centers[:, 0] - self.particle_box_size // 2 + self.pad
            x_rights = centers[:, 0] + self.particle_box_size // 2 + parity + self.pad
            y_lefts = centers[:, 1] - self.particle_box_size // 2 + self.pad
            y_rights = centers[:, 1] + self.particle_box_size // 2 + parity + self.pad
            # Subtract image from the micrograph using the particle's bounding box and center
            for p in range(self.particles_per_micrograph):
                clean_micrograph[m][
                    x_lefts[p] : x_rights[p], y_lefts[p] : y_rights[p]
                ] = (
                    clean_micrograph[m][
                        x_lefts[p] : x_rights[p], y_lefts[p] : y_rights[p]
                    ]
                    + images[p]
                )
        clean_micrograph = clean_micrograph[
            :,
            self.pad : self.micrograph_size + self.pad,
            self.pad : self.micrograph_size + self.pad,
        ]
        return Image(clean_micrograph)

    def get_micrograph_index(self, particle_index):
        """
        Using the global ID of the particle, returns the micrograph ID and the local particle ID.

        :param particle_id: Global ID of the particle.
        :return: The micrograph ID and the local ID of the particle.
        """
        if particle_index >= self.total_particle_count or particle_index < 0:
            raise RuntimeError("Index out of bounds.")
        return divmod(particle_index, self.particles_per_micrograph)

    def get_particle_indices(self, micrograph_index, particle_index=None):
        """
        Using the micrograph ID, returns every global particle ID from
        that micrograph. Returns specific global IDs if the local IDs
        are given.

        :param micrograph_index: ID of the micrograph.
        :param particle_index: Local ID of the particle.
        :return: The global ID of the particle.
        """
        if micrograph_index >= self.micrograph_count or micrograph_index < 0:
            raise RuntimeError("Index out of bounds for micrograph.")
        if particle_index is None:
            return np.arange(
                micrograph_index * self.particles_per_micrograph,
                (micrograph_index + 1) * self.particles_per_micrograph,
            )
        if particle_index >= self.particles_per_micrograph or particle_index < 0:
            raise RuntimeError("Index out of bounds for particle.")
        return micrograph_index * self.particles_per_micrograph + particle_index

    def save(self, path, name_prefix="micrograph", overwrite=True):
        """
        Save micrograph simulation to `path`.

        Currently saves micrographs to `.mrc`.
        Saves simulated centers, projection rotations, and CTF filter parameters
        when applicable.

        :param path: Directory to save data.
        :param name_prefix: Optional, name prefix string for micrograph files.
        :param overwrite: Optional, bool. Allow writing to existing directory,
            and overwriting existing files.
        :return: List of tuples [(`.mrc`, `.star`)..], compatible with CentersCoordinateSource.
        """

        _meta_fields = {
            "ctf": [
                "_rlnVoltage",
                "_rlnDefocusU",
                "_rlnDefocusV",
                "_rlnDefocusAngle",
                "_rlnSphericalAberration",
                "_rlnAmplitudeContrast",
            ],
            "rotations": ["_rlnAngleRot", "_rlnAngleTilt", "_rlnAnglePsi"],
        }

        output_mrc_filenames = super().save(
            path=path, name_prefix=name_prefix, overwrite=overwrite
        )

        output_star_filenames = []
        for m in range(self.micrograph_count):
            prefix = f"{name_prefix}_{m}"

            d = dict()

            # Particle to micrograph
            d["_rlnImageName"] = np.full(
                self.particles_per_micrograph, fill_value=""
            ).astype("object")
            d["_rlnImageName"][:] = [
                f"{j + 1:06}@{prefix}" for j in range(self.particles_per_micrograph)
            ]
            d["_rlnImageSize"] = np.full(
                self.particles_per_micrograph, fill_value=self.particle_box_size
            )

            # Particle centers
            y, x = zip(*self.centers[m])  # unzips
            d["_rlnCoordinateX"] = x
            d["_rlnCoordinateY"] = y

            # Projection rotations
            rots_metadata = self.simulation.get_metadata(
                metadata_fields=_meta_fields["rotations"],
                indices=self.get_particle_indices(m),
                as_dict=True,
            )

            # CTF
            ctf_metadata = dict()
            if self.simulation.unique_filters:
                ctf_metadata = self.simulation.get_metadata(
                    metadata_fields=_meta_fields["ctf"],
                    indices=self.get_particle_indices(m),
                    as_dict=True,
                )

            # Union dictionaries
            d = {**d, **rots_metadata, **ctf_metadata}

            # Write STAR file for this micrograph
            star_file_path = os.path.join(path, f"{prefix}.star")
            StarFile(blocks={"": d}).write(star_file_path)
            output_star_filenames.append(star_file_path)

        return list(zip(output_mrc_filenames, output_star_filenames))



"""
Wavelet-based approximate Earthmover's distance (EMD) for n-dimensional signals.

This code is based on the following paper:
    Sameer Shirdhonkar and David W. Jacobs.
    "Approximate earth mover’s distance in linear time."
    2008 IEEE Conference on Computer Vision and Pattern Recognition (CVPR).

More details are available in their technical report:
    CAR-TR-1025 CS-TR-4908 UMIACS-TR-2008-06.
"""

import warnings

import numpy as np
import pywt


def wemd_embed(arr, wavelet="coif3", level=None):
    """
    This function computes an embedding of Numpy arrays such that
    for non-negative arrays that sum to one, the L1 distance between the
    resulting embeddings is strongly equivalent to the Earthmover distance
    of the arrays.

    :param arr: Numpy array
    :param level: Decomposition level of the wavelets.
    Larger levels yield more coefficients and more accurate results.
    If no level is given, we take the the log2 of the side-length of the domain.
    :param wavelet: Either the name of a wavelet supported by PyWavelets
    (e.g. 'coif3', 'sym3', 'sym5', etc.) or a pywt.Wavelet object
    See https://pywavelets.readthedocs.io/en/latest/ref/wavelets.html#built-in-wavelets-wavelist
    The default is 'coif3', because it seems to work well empirically.
    :returns: One-dimensional numpy array containing weighted details coefficients.
    """
    dimension = arr.ndim

    if level is None:
        level = int(np.ceil(np.log2(max(arr.shape)))) + 1

    # Using wavedecn with the default level creates this boundary effects warning.
    # However, this doesn't seem to be a cause for concern.
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="Level value of .* is too high:"
            " all coefficients will experience boundary effects.",
        )
        arrdwt = pywt.wavedecn(arr, wavelet, mode="zero", level=level)

    detail_coefs = arrdwt[1:]
    assert len(detail_coefs) == level

    weighted_coefs = []
    for j, details_level_j in enumerate(detail_coefs):
        multiplier = 2 ** ((level - 1 - j) * (1 + (dimension / 2.0)))
        for coefs in details_level_j.values():
            weighted_coefs.append(multiplier * coefs.flatten())

    return np.concatenate(weighted_coefs)


def wemd_norm(arr, wavelet="coif3", level=None):
    """
    Wavelet-based norm used to approximate the Earthmover's distance between
    mass distributions specified as Numpy arrays (typically images or volumes).

    :param arr: Numpy array of the difference between the two mass distributions.
    :param level: Decomposition level of the wavelets.
    Larger levels yield more coefficients and more accurate results.
    If no level is given, we take the the log2 of the side-length of the domain.
    Larger levels yield more coefficients and more accurate results
    :param wavelet: Either the name of a wavelet supported by PyWavelets
    (e.g. 'coif3', 'sym3', 'sym5', etc.) or a pywt.Wavelet object
    See https://pywavelets.readthedocs.io/en/latest/ref/wavelets.html#built-in-wavelets-wavelist
    The default is 'coif3', because it seems to work well empirically.
    :return: Approximated Earthmover's Distance
    """

    coefs = wemd_embed(arr, wavelet, level)
    return np.linalg.norm(coefs, ord=1)



import copy
import logging

import numpy as np
from scipy.linalg import eigh, qr
from sklearn.metrics import adjusted_rand_score

from aspire.image import Image
from aspire.noise import NoiseAdder
from aspire.source import ImageSource
from aspire.source.image import _ImageAccessor
from aspire.utils import (
    acorr,
    ainner,
    anorm,
    make_symmat,
    uniform_random_angles,
    vecmat_to_volmat,
)
from aspire.utils.random import rand, randi, randn
from aspire.volume import AsymmetricVolume, Volume

logger = logging.getLogger(__name__)


class Simulation(ImageSource):
    """
    A `Simulation` represents a synthetic dataset of realistic cryo-EM images with corresponding
    `metadata`. The images are generated via projections of a supplied `Volume` object, `vols`, over
    orientations define by the Euler angles, `angles`. Various types of corruption, such as noise and
    CTF effects, can be added to the images by supplying a `Filter` object to the `noise_filter` or
    `unique_filters` arguments.
    """

    def __init__(
        self,
        L=None,
        n=1024,
        vols=None,
        states=None,
        unique_filters=None,
        filter_indices=None,
        offsets=None,
        amplitudes=None,
        dtype=None,
        C=2,
        angles=None,
        seed=0,
        memory=None,
        noise_adder=None,
        symmetry_group=None,
    ):
        """
        A `Simulation` object that supplies images along with other parameters for image manipulation.

        :param L: Resolution of projection images (integer). Default is 8.
            If a `Volume` is provided `L` and `vols.resolution` must agree.
        :param n: The number of images to generate (integer).
        :param vols: A `Volume` object representing a stack of volumes.
            Default is generated with `volume.volume_synthesis.AsymmetricVolume`.
        :param states: A 1d array of n integers in the interval [0, C). The i'th integer indicates
            the volume stack index used to produce the i'th projection image. Default is a random set.
        :param unique_filters: A list of Filter objects to be applied to projection images.
        :param filter_indices: A 1d array of n integers indicating the `unique_filter` indices associated
            with each image. Default is a random set of filter indices, .ie the filters from `unique_filters`
            are randomly assigned to the stack of images.
        :param offsets: A n-by-2 array of coordinates to offset the images. Default is a normally
            distributed set of offsets. Set `offsets = 0` to disable offsets.
        :param amplitude: A 1d array of n amplitudes to scale the projection images. Default is
            a random set in the interval [2/3, 3/2]. Set `amplitude = 1` to disable amplitudes.
        :param dtype: dtype for the Simulation
        :param C: Number of Volumes used to generate projection images. The default is C=2.
            If a `Volume` object is provided this parameter is overridden and `self.C` = `self.vols.n_vols`.
        :param angles: A n-by-3 array of Euler angles for use in projection. Default is a random set.
        :param seed: Random seed.
        :param memory: str or None. The path of the base directory to use as a data store or None.
            If None is given, no caching is performed.
        :param noise_adder: Optionally append instance of `NoiseAdder`
            to generation pipeline.
        :param symmetry_group: A SymmetryGroup instance or string indicating symmetry of the molecule.

        :return: A Simulation object.
        """

        self.seed = seed

        # If a Volume is not provided we default to the legacy Gaussian blob volume.
        # If a Simulation resolution or dtype is not provided, we default to L=8 and np.float32.
        if vols is None:
            self.vols = AsymmetricVolume(
                L=L or 8,
                C=C,
                seed=self.seed,
                dtype=dtype or np.float32,
            ).generate()
        else:
            if dtype is not None and vols.dtype != dtype:
                raise RuntimeError(
                    f"Explicit {self.__class__.__name__} dtype {dtype}"
                    f" does not match provided vols.dtype {vols.dtype}."
                    " Please change the Volume using `astype`"
                    " or the Simuation `dtype` argument."
                )
            # Assign the explicitly provided volume.
            self.vols = vols

        if not isinstance(self.vols, Volume):
            raise RuntimeError("`vols` should be a Volume instance or `None`.")

        if symmetry_group and (str(symmetry_group) != str(self.vols.symmetry_group)):
            logger.warning(
                f"Overriding {str(self.vols.symmetry_group)} symmetry group inherited "
                f"from `vols`, with user provided symmetry group: {str(symmetry_group)}."
            )
        symmetry_group = symmetry_group or self.vols.symmetry_group

        # Infer the details from volume when possible.
        super().__init__(
            L=self.vols.resolution,
            n=n,
            dtype=self.vols.dtype,
            memory=memory,
            symmetry_group=symmetry_group,
        )

        # If a user provides both `L` and `vols`, resolution should match.
        if L is not None and L != self.L:
            raise RuntimeError(
                f"Simulation must have the same resolution as the provided Volume."
                f" Provided vols.resolution = {self.vols.resolution} and L = {L}."
            )

        # We need to keep track of the original resolution we were initialized with,
        # to be able to generate projections of volumes later, when we are asked to supply images.
        self._original_L = self.L

        if offsets is None:
            offsets = self.L / 16 * randn(2, n, seed=seed).astype(dtype).T

        if amplitudes is None:
            min_, max_ = 2.0 / 3, 3.0 / 2
            amplitudes = min_ + rand(n, seed=seed).astype(dtype) * (max_ - min_)

        self.C = self.vols.n_vols

        if states is None:
            states = randi(self.C, n, seed=seed)
        self.states = states

        if angles is None:
            angles = uniform_random_angles(n, seed=seed, dtype=self.dtype)
        self.angles = angles

        if unique_filters is None:
            unique_filters = []
        self.unique_filters = unique_filters
        # sim_filters must be a deep copy so that it is not changed
        # when unique_filters is changed
        self.sim_filters = copy.deepcopy(unique_filters)

        # Create filter indices and fill the metadata based on unique filters
        if unique_filters:
            if filter_indices is None:
                filter_indices = randi(len(unique_filters), n, seed=seed) - 1
            self._populate_ctf_metadata(filter_indices)
            self.filter_indices = filter_indices
        else:
            self.filter_indices = np.zeros(n, dtype=int)

        self.offsets = offsets
        self.amplitudes = amplitudes

        self._projections_accessor = _ImageAccessor(self._projections, self.n)
        self._clean_images_accessor = _ImageAccessor(self._clean_images, self.n)

        # If a user prescribed NoiseAdder.from_snr(...),
        #   noise_adder will be a function returning a completed class.
        # Note the delayed eval may attempt to use self.*_accessors above.
        if noise_adder is not None:
            logger.info(f"Appending {noise_adder} to generation pipeline")
            # If we need to calculate signal_power from Simulation,
            # do so now and assign it to complete the Filter.
            if getattr(noise_adder, "requires_signal_power", False):
                noise_adder.signal_power = self.true_signal_power()

            # At this point we should have a fully baked NoiseAdder
            if not isinstance(noise_adder, NoiseAdder):
                raise RuntimeError("`noise_adder` should be subclass of NoiseAdder")

        self.noise_adder = noise_adder

        # Any further operations should not mutate this instance.
        self._mutable = False

    def _populate_ctf_metadata(self, filter_indices):
        # Since we are not reading from a starfile, we must construct
        # metadata based on the CTF filters by hand and set the values
        # for these columns
        #
        # class attributes of CTFFilter:
        CTFFilter_attributes = (
            "voltage",
            "defocus_u",
            "defocus_v",
            "defocus_ang",
            "Cs",
            "alpha",
        )
        # get the CTF parameters, if they exist, for each filter
        # and for each image (indexed by filter_indices)
        filter_values = np.zeros((len(filter_indices), len(CTFFilter_attributes)))
        for i, filt in enumerate(self.unique_filters):
            filter_values[filter_indices == i] = [
                getattr(filt, att, np.nan) for att in CTFFilter_attributes
            ]
        # set the corresponding Relion metadata values that we would expect
        # from a STAR file
        self.set_metadata(
            [
                "_rlnVoltage",
                "_rlnDefocusU",
                "_rlnDefocusV",
                "_rlnDefocusAngle",
                "_rlnSphericalAberration",
                "_rlnAmplitudeContrast",
            ],
            filter_values,
        )

    @property
    def projections(self):
        """
        Return projections of generated volumes, without applying filters/shifts/amplitudes/noise

        :param start: start index (0-indexed) of the start image to return
        :param num: Number of images to return. If None, *all* images are returned.
        :param indices: A numpy array of image indices. If specified, start and num are ignored.
        :return: An Image instance.
        """
        return self._projections_accessor

    def _projections(self, indices):
        """
        Accesses and returns projections as an `Image` instance. Called by self._projections_accessor
        """
        im = np.zeros(
            (len(indices), self._original_L, self._original_L), dtype=self.dtype
        )

        states = self.states[indices]
        unique_states = np.unique(states)
        for k in unique_states:
            idx_k = np.where(states == k)[0]
            rot = self.rotations[indices[idx_k], :, :]

            im_k = self.vols[k - 1].project(rot_matrices=rot)
            im[idx_k, :, :] = im_k.asnumpy()

        return Image(im)

    @property
    def clean_images(self):
        """
        Return projections with filters/shifts/amplitudes applied, but without noise.
        Subscriptable property.
        """
        return self._clean_images_accessor

    def _clean_images(self, indices):
        return self._images(indices, clean_images=True)

    def _images(self, indices, clean_images=False):
        """
        Returns particle images when accessed via the `ImageSource.images` property.
        :param indices: A 1-D NumPy array of integer indices.
        :param clean_images: Only used internally, toggled on when `clean_images` requested.
             Will skip accessing cache, skip noise, while applying CTF to projections.
        :return: An `Image` object.
        """
        # check for cached images first
        if not clean_images and self._cached_im is not None:
            logger.debug("Loading images from cache")
            return self.generation_pipeline.forward(self._cached_im[indices], indices)

        im = self.projections[indices]

        # apply original CTF distortion to image
        im = self._apply_sim_filters(im, indices)

        im = im.shift(self.offsets[indices, :])

        im *= self.amplitudes[indices].reshape(len(indices), 1, 1).astype(self.dtype)

        if not clean_images and self.noise_adder is not None:
            im = self.noise_adder.forward(im, indices=indices)

        # Finally, apply transforms to resulting Image
        return self.generation_pipeline.forward(im, indices)

    def _apply_sim_filters(self, im, indices):
        return self._apply_filters(
            im,
            self.sim_filters,
            self.filter_indices[indices],
        )

    def vol_coords(self, mean_vol=None, eig_vols=None):
        """
        Coordinates of simulation volumes in a given basis

        :param mean_vol: A mean volume in the form of a Volume Instance (default `mean_true`).
        :param eig_vols: A set of k volumes in a Volume instance (default `eigs`).
        :return:
        """
        if mean_vol is None:
            mean_vol = self.mean_true()
        if eig_vols is None:
            eig_vols = self.eigs()[0]

        assert isinstance(mean_vol, Volume)
        assert isinstance(eig_vols, Volume)

        vols = self.vols - mean_vol  # note, broadcast

        V = vols.to_vec()
        EV = eig_vols.to_vec()

        coords = EV @ V.T

        res = vols - Volume.from_vec(coords.T @ EV)
        res_norms = anorm(res.asnumpy(), (1, 2, 3))
        res_inners = mean_vol.to_vec() @ res.to_vec().T

        return coords.squeeze(), res_norms, res_inners

    def true_signal_power(self, *args, **kwargs):
        """
        Estimate the signal power of `clean_images`.

        For usage, see `ImageSource.estimate_signal_power`.

        :returns: Estimated signal power of `clean_images`
        """

        # Note, in the future we can do something more clever here,
        # perhaps starting with the simulation Volume.  For now we
        # share code with ImageSource, so the method is at least
        # identical, up to using `clean_images`.  The method is also
        # the same as NoiseEstimator, up to ignoring the first moment
        # and a few optional parameters.
        kwargs["image_accessor"] = self.clean_images
        return self.estimate_signal_power(*args, **kwargs)

    def mean_true(self):
        return Volume(np.mean(self.vols, 0))

    def covar_true(self):
        eigs_true, lamdbas_true = self.eigs()
        eigs_true = eigs_true.T.to_vec()

        covar_true = eigs_true.T @ lamdbas_true @ eigs_true
        covar_true = vecmat_to_volmat(covar_true)

        return covar_true

    def eigs(self):
        """
        Eigendecomposition of volume covariance matrix of simulation
        :return: A 2-tuple:
            eigs_true: The eigenvectors of the volume covariance matrix in the form of Volume instance.
            lambdas_true: The eigenvalues of the covariance matrix in the form of a (C-1)-by-(C-1) diagonal matrix.
        """
        C = self.C
        vols_c = self.vols - self.mean_true()

        p = np.ones(C) / C
        # RCOPT, we may be able to do better here if we dig in.
        Q, R = qr(vols_c.to_vec().T, mode="economic")

        # Rank is at most C-1, so remove last vector
        Q = Q[:, :-1]
        R = R[:-1, :]

        w, v = eigh(make_symmat(R @ np.diag(p) @ R.T))
        eigs_true = Volume.from_vec((Q @ v).T)

        # Arrange in descending order (flip column order in eigenvector matrix)
        w = w[::-1]
        eigs_true = Volume(eigs_true.asnumpy()[::-1])

        return eigs_true, np.diag(w)

    # TODO: Too many eval_* methods doing similar things - encapsulate somehow?

    def eval_mean(self, mean_est):
        mean_true = self.mean_true()
        return self.eval_vol(mean_true, mean_est)

    def eval_vol(self, vol_true, vol_est):
        norm_true = anorm(vol_true)

        err = anorm(vol_true - vol_est)
        rel_err = err / norm_true
        # RCOPT
        corr = acorr(vol_true.asnumpy(), vol_est.asnumpy())

        return {"err": err, "rel_err": rel_err, "corr": corr}

    def eval_covar(self, covar_est):
        covar_true = self.covar_true()
        return self.eval_volmat(covar_true, covar_est)

    def eval_volmat(self, volmat_true, volmat_est):
        """
        Evaluate volume matrix estimation accuracy

        :param volmat_true: The true volume matrices in the form of an L-by-L-by-L-by-L-by-L-by-L-by-K array.
        :param volmat_est: The estimated volume matrices in the same form.
        :return:
        """
        norm_true = anorm(volmat_true)

        err = anorm(volmat_true - volmat_est)
        rel_err = err / norm_true
        corr = acorr(volmat_true, volmat_est)

        return {"err": err, "rel_err": rel_err, "corr": corr}

    def eval_eigs(self, eigs_est, lambdas_est):
        """
        Evaluate covariance eigendecomposition accuracy

        :param eigs_est: The estimated volume eigenvectors in an L-by-L-by-L-by-K array.
        :param lambdas_est: The estimated eigenvalues in a K-by-K diagonal matrix (default `diag(ones(K, 1))`).
        :return:
        """
        eigs_true, lambdas_true = self.eigs()

        B = eigs_est.to_vec() @ eigs_true.to_vec().T
        norm_true = anorm(lambdas_true)
        norm_est = anorm(lambdas_est)

        inner = ainner(B @ lambdas_true, lambdas_est @ B)
        err = np.sqrt(norm_true**2 + norm_est**2 - 2 * inner)
        rel_err = err / norm_true
        corr = inner / (norm_true * norm_est)

        # TODO: Determine Principal Angles and return as a dict value

        return {"err": err, "rel_err": rel_err, "corr": corr}

    def eval_clustering(self, vol_idx):
        """
        Evaluate clustering estimation using an adjusted Rand score.

        :param vol_idx: Indexes of the volumes determined (0-indexed)
        :return: Accuracy [-0.5, 1] in terms of proportion of correctly assigned labels.
            Identical clusters (up to a permutation) have a score of 1, random labeling
            will be close to 0, and discordant clusterings will be negative.
        """
        assert (
            len(vol_idx) == self.n
        ), f"Need {self.n} vol indexes to evaluate clustering"
        # Remember that `states` is 1-indexed while vol_idx is 0-indexed.

        return adjusted_rand_score(self.states - 1, vol_idx)

    def eval_coords(self, mean_vol, eig_vols, coords_est):
        """
        Evaluate coordinate estimation

        :param mean_vol: A mean volume in the form of a Volume instance.
        :param eig_vols: A set of eigenvolumes in an Volume instance.
        :param coords_est: The estimated coordinates in the affine space defined centered at `mean_vol` and spanned
            by `eig_vols`.
        :return:
        """
        assert isinstance(mean_vol, Volume)
        assert isinstance(eig_vols, Volume)
        coords_true, res_norms, res_inners = self.vol_coords(mean_vol, eig_vols)

        # 0-indexed states vector
        states = self.states - 1
        coords_true = coords_true[states]
        res_norms = res_norms[states]
        res_inners = res_inners[:, states]

        mean_eigs_inners = (mean_vol.to_vec() @ eig_vols.to_vec().T).item()
        coords_err = coords_true - coords_est

        err = np.hypot(res_norms, coords_err)

        mean_vol_norm2 = anorm(mean_vol) ** 2
        norm_true = np.sqrt(
            coords_true**2
            + mean_vol_norm2
            + 2 * res_inners
            + 2 * mean_eigs_inners * coords_true
        )
        norm_true = np.hypot(res_norms, norm_true)

        rel_err = err / norm_true
        inner = (
            mean_vol_norm2
            + mean_eigs_inners * (coords_true + coords_est)
            + coords_true * coords_est
            + res_inners
        )
        norm_est = np.sqrt(
            coords_est**2 + mean_vol_norm2 + 2 * mean_eigs_inners * coords_est
        )

        corr = inner / (norm_true * norm_est)

        return {"err": err, "rel_err": rel_err, "corr": corr}

    def true_snr(self, *args, **kwargs):
        """
        Compute SNR using `true_signal_power` and the noise power known to simulation.

        See Simulation.true_signal_power() for parameters.
        """
        # For clean images return infinite SNR.
        # Note, relationship with CTF and other sim corruptions still isn't clear to me...
        if self.noise_adder is None or self.noise_adder.noise_var == 0:
            return np.inf

        # For SNR of Simulations, use the theoretical noise variance
        # known from the noise_adder instead of deriving from PSD.
        noise_power = self.noise_adder.noise_var
        signal_power = self.true_signal_power(*args, **kwargs)
        return signal_power / noise_power



import logging
import platform
import struct
import subprocess
import sys
import traceback


def handle_exception(exc_type, exc_value, exc_traceback):
    """
    Handle any top-level unhandled exception.
    Tries to gather and log additional context information,
    then re-raises.

    :param exc_type: Exception type object
    :param exc_value: Exception value object (an instance of type exc_type)
    :param exc_traceback: The Traceback object associated with exc_value (also available as exc_value.__traceback__)
    :return: On return, useful diagnostic information has been logged, and the exception re-raised.
    """

    # Are we explicitly/interactively killing a run? Just do it.
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    from aspire.utils import get_full_version

    lines = list()

    lines.append(f"Application version: {get_full_version()}")
    lines.append(f"Platform: {platform.platform()}")
    lines.append(f"Python version: {sys.version}")
    lines.append(f'Python 32/64 bit: {8 * struct.calcsize("P")}')

    lines.append("conda list output:")
    try:
        lines.extend(
            subprocess.check_output(["conda", "list"], stderr=subprocess.STDOUT)
            .decode("utf8")
            .split("\n")
        )
    except Exception:  # nopep8  # noqa: E722
        pass

    lines.append("pip freeze output:")
    try:
        lines.extend(
            subprocess.check_output(["pip", "freeze"], stderr=subprocess.STDOUT)
            .decode("utf8")
            .split("\n")
        )
    except Exception:  # nopep8  # noqa: E722
        pass

    # Walk through all traceback objects (oldest call -> most recent call), capturing frame/local variable information.
    lines.append("Exception Details (most recent call last)")
    frame_generator = traceback.walk_tb(exc_traceback)

    try:
        stack_summary = traceback.StackSummary.extract(
            frame_generator, capture_locals=True
        )
    except Exception:  # nopep8  # noqa: E722
        # The above code, while more informative, doesn't always work.
        # When it doesn't try something simpler.
        stack_summary = traceback.StackSummary.extract(
            frame_generator, capture_locals=False
        )

    frame_strings = stack_summary.format()
    for s in frame_strings:
        lines.extend(s.split("\n"))

    try:
        with open("aspire.err.log", "w") as f:
            f.write("\n".join(lines) + "\n")
    except Exception:  # nopep8  # noqa: E722
        pass

    try:
        # send to logger
        logging.critical(
            f"{exc_value}\nTraceback:\n"
            f'{"".join(traceback.format_tb(exc_traceback))}'
        )
        # re-raise the exception we got for the caller.
        raise exc_value
    finally:
        # cleanup - see https://cosmicpercolator.com/2016/01/13/exception-leaks-in-python-2-and-3/
        del exc_value, exc_traceback


# Useful Exception classes
class AspireException(Exception):
    pass


class WrongInput(AspireException):
    pass


class DimensionsIncompatible(AspireException):
    pass


class ErrorTooBig(AspireException):
    pass


class UnknownFormat(AspireException):
    pass



"""
Define related utility functions for Fourier–Bessel (2D), Spherical Fourier–Bessel (3D) and
prolate spheroidal wave function (PSWF) objects.
"""

import logging

import numpy as np
from numpy import diff, exp, log, pi
from numpy.polynomial.legendre import leggauss
from scipy.special import jn, jv, sph_harm

from aspire.utils import grid_2d, grid_3d

logger = logging.getLogger(__name__)


def check_besselj_zeros(nu, z):
    """
    Sanity-check a sequence of estimated zeros of the Bessel function with order `nu`.

    :param nu: The real number order of the Bessel function.
    :param z: (Array-like) A sequence of postulated zeros.
    :return result: True or False.
    """
    # Compute first and second order differences of the sequence of zeros
    dz = np.diff(z)
    ddz = np.diff(dz)

    # Check criteria for acceptable zeros
    result = True
    # Real roots
    result = result and all(np.isreal(z))
    # All roots should be > 0, check first of increasing sequence
    result = result and z[0] > 0
    # Spacing between zeros is greater than 3
    result = result and all(dz > 3)

    # Second order differences should be zero or just barely increasing to
    # within 16x machine precision.
    if nu >= 0.5:
        result = result and all(ddz < 16 * np.spacing(z[1:-1]))
    # For nu < 0.5 the spacing will be slightly decreasing, so flip the sign
    else:
        result = result and all(ddz > -16 * np.spacing(z[1:-1]))

    return result


def besselj_newton(nu, z0, max_iter=10):
    """
    Uses the Newton-Raphson method to compute the zero(s) of the
    Bessel function with order `nu` with initial guess(es) `z0`.

    :param nu: The real number order of the Bessel function.
    :param z0: (Array-like) The initial guess(es) for the root-finding algorithm.
    :param max_iter: Maximum number of iterations for Newton-Raphson
        (default: 10).
    :return z: (Array-like) The estimated root(s).
    """
    z = z0

    # Factor worse than machine precision
    c = 8

    for i in range(max_iter):
        # Calculate values and derivatives at z
        f = jv(nu, z)
        fp = jv(nu - 1, z) - nu * f / z

        # Update zeros
        dz = -f / fp
        z = z + dz

        # Check for convergence
        if all(np.abs(dz) < c * np.spacing(z)):
            break

        # If we're not converging yet, start relaxing convergence criterion
        if i >= 6:
            c *= 2

    return z


def sph_bessel(ell, r):
    """
    Compute spherical Bessel function values.

    :param ell: The order of the spherical Bessel function.
    :param r: The coordinates where the function is to be evaluated.
    :return: The value of j_ell at r.
    """
    scalar = np.isscalar(r)
    len_r = 1 if scalar else len(r)

    j = np.zeros(len_r)
    j[r == 0] = 1 if ell == 0 else 0

    r_mask = r != 0
    j[r_mask] = np.sqrt(pi / (2 * r[r_mask])) * jv(ell + 0.5, r[r_mask])

    if scalar:
        j = j.item()

    return j


def norm_assoc_legendre(j, m, x):
    """
    Evaluate the normalized associated Legendre polynomial

    :param j: The order of the associated Legendre polynomial, must satisfy |m| < j.
    :param m: The degree of the associated Legendre polynomial, must satisfy |m| < j.
    :param x: An array of values between -1 and +1 on which to evaluate.
    :return: The normalized associated Legendre polynomial evaluated at corresponding x.

    """

    # For negative m, flip sign and use the symmetry identity.
    # In the rest, we assume that m is non-negative.
    if m < 0:
        m = -m
        px = (-1) ** m * norm_assoc_legendre(j, m, x)
        px *= (-1) ** m
        return px

    # Initialize the recurrence at (m, m) and (m, m+1).
    p0 = (
        (-1) ** m
        * np.sqrt(
            (2 * m + 1)
            / 2
            * np.prod(np.arange(2 * m - 1, 0, -2) / np.arange(2 * m, 0, -2))
        )
        * (1 - x * x) ** (m / 2)
    )

    p1 = x * np.sqrt(2 * m + 3) * p0

    # If these are the desired indices, return these initial values.
    if j == m:
        px = p0
    elif j == m + 1:
        px = p1
    else:
        # Fixing m, work our way up from (m, m+1) to (m, j).
        for n in range(m + 1, j):
            px = np.sqrt((2 * n + 3) / ((n + 1 + m) * (n + 1 - m))) * (
                np.sqrt(2 * n + 1) * x * p1
                - np.sqrt((n + m) * (n - m) / (2 * n - 1)) * p0
            )
            p0 = p1
            p1 = px

    return px


def real_sph_harmonic(j, m, theta, phi):
    """
    Evaluate a real spherical harmonic

    :param j: The order of the spherical harmonic. These must satisfy |m| < j.
    :param m: The degree of the spherical harmonic. These must satisfy |m| < j.
    :param theta: The spherical coordinates of the points at which we want to evaluate the real spherical harmonic.
        `theta` is the latitude between 0 and pi
    :param phi: The spherical coordinates of the points at which we want to evaluate the real spherical harmonic.
        `phi` is the longitude, between 0 and 2*pi
    :return: The real spherical harmonics evaluated at the points (theta, phi).
    """
    abs_m = abs(m)

    y = sph_harm(abs_m, j, phi, theta)
    if m < 0:
        y = np.sqrt(2) * np.imag(y)
    elif m > 0:
        y = np.sqrt(2) * np.real(y)
    else:
        y = np.real(y)

    return y


def besselj_zeros(nu, k):
    """
    Finds the first `k` zeros of the Bessel function of order `nu`, i.e. J_nu.
    Adapted from "zerobess.m" by Jonas Lundgren <splinefit@gmail.com>

    :param nu: The real number order of the Bessel function (must be positive and <1e7).
    :param k: The number of zeros to return (must be >= 3).
    :return z: A 1D NumPy array of the first `k` zeros.
    """
    assert k >= 3, "k must be >= 3"
    assert 0 <= nu <= 1e7, "nu must be between 0 and 1e7"

    z = np.zeros(k)

    # Guess first zeros using powers of nu
    c0 = np.array(
        [
            [0.1701, -0.6563, 1.0355, 1.8558],
            [0.1608, -1.0189, 3.1348, 3.2447],
            [-0.2005, -1.2542, 5.7249, 4.3817],
        ]
    )
    z0 = nu + c0 @ ((nu + 1) ** np.array([[-1, -2 / 3, -1 / 3, 1 / 3]]).T)

    # refine guesses
    z[:3] = besselj_newton(nu, z0).squeeze()

    n = 3
    j = 2
    err_tol = 5e-3

    # Estimate further zeros iteratively using spacing of last three zeros so far
    while n < k:
        j = min(j, k - n)

        # Use last 3 zeros to predict spacing for next j zeros
        r = diff(z[n - 3 : n]) - pi
        if (r[0] * r[1]) > 0 and (r[0] / r[1]) > 1:
            p = log(r[0] / r[1]) / log(1 - 1 / (n - 1))
            t = np.array(np.arange(1, j + 1), ndmin=2).T / (n - 1)
            dz = pi + r[1] * exp(p * log(1 + t))
        else:
            dz = pi * np.ones((j, 1))

        # Guess and refine
        z0 = z[n - 1] + np.cumsum(dz)
        z[n : n + j] = besselj_newton(nu, z0)

        # Check to see that the sequence of zeros makes sense
        assert check_besselj_zeros(
            nu, z[n - 2 : n + j]
        ), "Unable to properly estimate Bessel function zeros."

        # Check how far off we are
        err = (z[n : n + j] - z0) / np.diff(z[n - 1 : n + j])

        n = n + j
        if max(abs(err)) < err_tol:
            # Predictions were close enough, double number of zeros
            j *= 2
        else:
            # Some predictions were off, set to double the number of good predictions
            j = 2 * (np.where(abs(err) >= err_tol)[0][0] + 1)

    return z


def all_besselj_zeros(ell, r):
    """
    Compute the zeros of the order `ell` Bessel function which are less than `r`.

    :param ell: The real number order of the Bessel function.
    :param r: The upper bound for zeros returned.
    :return n, r0: The number of zeros and the zeros themselves
        as a NumPy array.
    """
    k = 4
    # get the first 4 zeros
    r0 = besselj_zeros(ell, k)
    while all(r0 < r):
        # increase the number of zeros sought
        # until one of the zeros is greater than `r`
        k *= 2
        r0 = besselj_zeros(ell, k)
    r0 = r0[r0 < r]
    # return the number of zeros and the zeros themselves
    return len(r0), r0


def unique_coords_nd(N, ndim, shifted=False, normalized=True, dtype=np.float32):
    """
    Generate unique polar coordinates from 2D or 3D rectangular coordinates.

    :param N: length size of a square or cube.
    :param ndim: number of dimension, 2 or 3.
    :param shifted: shifted half pixel or not for odd N.
    :param normalized: normalize the grid or not.
    :return: The unique polar coordinates in 2D or 3D
    """
    assert ndim in (
        2,
        3,
    ), "Only two- or three-dimensional basis functions are supported."
    assert N > 0, "Number of grid points should be greater than 0."

    if ndim == 2:
        grid = grid_2d(
            N, shifted=shifted, normalized=normalized, indexing="yx", dtype=dtype
        )
        mask = grid["r"] <= 1

        # Minor differences in r/theta/phi values are unimportant for the purpose
        # of this function, so round off before proceeding

        r = grid["r"][mask].round(5)
        phi = grid["phi"][mask].round(5)

        r_unique, r_idx = np.unique(r, return_inverse=True)
        ang_unique, ang_idx = np.unique(phi, return_inverse=True)

    else:
        grid = grid_3d(
            N, shifted=shifted, normalized=normalized, indexing="zyx", dtype=dtype
        )
        mask = grid["r"] <= 1

        # Minor differences in r/theta/phi values are unimportant for the purpose of this function,
        # so we round off before proceeding.

        r = grid["r"][mask].round(5)
        theta = grid["theta"][mask].round(5)
        phi = grid["phi"][mask].round(5)

        r_unique, r_idx = np.unique(r, return_inverse=True)
        ang_unique, ang_idx = np.unique(
            np.vstack([theta, phi]), axis=1, return_inverse=True
        )

    return {
        "r_unique": r_unique,
        "ang_unique": ang_unique,
        "r_idx": r_idx,
        "ang_idx": ang_idx,
        "mask": mask,
    }


def lgwt(ndeg, a, b, dtype=np.float32):
    """
    Compute Legendre-Gauss quadrature

    Generates the Legendre-Gauss nodes and weights on an interval
    [a, b] with truncation order of ndeg for computing definite integrals
    using Legendre-Gauss quadrature.
    Suppose you have a continuous function f(x) which is defined on [a, b]
    which you can evaluate at any x in [a, b]. Simply evaluate it at all of
    the values contained in the x vector to obtain a vector f, then compute
    the definite integral using sum(f.*w);

    This is a 2rapper for numpy.polynomial leggauss which outputs only in the
    range of (-1, 1).

    :param ndeg: truncation order, that is, the number of nodes.
    :param a, b: The endpoints of the interval over which the quadrature is defined.
    :return x, w: The quadrature nodes and weights.
    """

    x, w = leggauss(ndeg)
    scale_factor = (b - a) / 2
    shift = (a + b) / 2
    x = scale_factor * x + shift
    w = scale_factor * w

    return x.astype(dtype), w.astype(dtype)


def d_decay_approx_fun(a, b, c, d):
    return np.square(c) / (16 * (np.square(d) + d * (2 * b + a + 1)) - np.square(c))


def p_n(n, alpha, beta, x):
    """
    The first n jacobi polynomial of x as defined in Yoel's PSWF paper, eq (2), page 6

    :param n: int, > 0
        Number of polynomials to compute
    :param alpha: float, > -1
    :param beta: float, > -1
    :param x: (m,) ndarray
    :return: v: (m, n + 1) ndarray
        v[:, i] = P^{(alpha, beta)}_n(x) as defined in the paper
    """
    m = len(x)
    if n < 0:
        return np.array([])
    if n == 0:
        return np.ones(m)
    x = x.reshape((m,))
    v = np.zeros((m, n + 1))
    alpha_p_beta = alpha + beta
    alpha_m_beta = alpha - beta
    v[:, 0] = 1
    v[:, 1] = (1 + 0.5 * alpha_p_beta) * x + 0.5 * alpha_m_beta

    for i in range(2, n + 1):
        c1 = 2 * i * (i + alpha_p_beta) * (2 * i + alpha_p_beta - 2)
        c2 = (
            (2 * i + alpha_p_beta)
            * (2 * i + alpha_p_beta - 1)
            * (2 * i + alpha_p_beta - 2)
        )
        c3 = (2 * i + alpha_p_beta - 1) * alpha_p_beta * alpha_m_beta
        c4 = -2 * (i - 1 + alpha) * (i - 1 + beta) * (2 * i + alpha_p_beta)
        v[:, i] = ((c3 + c2 * x) * v[:, i - 1] + c4 * v[:, i - 2]) / c1
    return v


def t_x_mat(x, n, j, approx_length):
    a = np.power(x, n + 0.5)
    b = np.sqrt(2 * (2 * j + n + 1))
    c = p_n(approx_length - 1, n, 0, 1 - 2 * np.square(x))
    return np.einsum("i,j,ij->ij", a, b, c)


def t_x_mat_dot(x, n, j, approx_length):
    #  x need to be a matrix instead of a vector in t_x_mat
    return np.power(x, n + 0.5).dot(np.sqrt(2 * (2 * j + n + 1))) * p_n(
        approx_length - 1, n, 0, 1 - 2 * np.square(x)
    )


def t_x_derivative_mat(t1, t2, x, big_n, range_array, approx_length):
    return -2 * (big_n + range_array + 1) * np.outer(
        np.power(x, big_n + 1.5), t2
    ) * np.column_stack(
        (np.zeros(len(x)), p_n(approx_length - 2, big_n + 1, 1, t1))
    ) + (
        big_n + 0.5
    ) * np.outer(
        np.power(x, big_n - 0.5), t2
    ) * p_n(
        approx_length - 1, big_n, 0, t1
    )


def t_radial_part_mat(x, n, j, m):
    a = np.power(x, n)
    b = np.sqrt(2 * (2 * j + n + 1))
    c = p_n(m - 1, n, 0, 1 - 2 * np.square(x))
    return np.einsum("i,j,ij->ij", a, b, c)


def k_operator(nu, x):
    return jn(nu, x) * np.sqrt(x)



from .symmetry_groups import (
    CnSymmetryGroup,
    DnSymmetryGroup,
    IdentitySymmetryGroup,
    OSymmetryGroup,
    SymmetryGroup,
    TSymmetryGroup,
)
from .volume import Volume, qr_vols_forward, rotated_grids, rotated_grids_3d

from .volume_synthesis import (  # isort:skip
    TSymmetricVolume,
    OSymmetricVolume,
    CnSymmetricVolume,
    DnSymmetricVolume,
    AsymmetricVolume,
    LegacyVolume,
)



import logging

import numpy as np
import pycuda.autoinit  # noqa: F401
import pycuda.driver as cuda  # noqa: F401
import pycuda.gpuarray as gpuarray  # noqa: F401
from cufinufft import cufinufft

from aspire.nufft import Plan
from aspire.utils import complex_type

logger = logging.getLogger(__name__)


class CufinufftPlan(Plan):
    def __init__(self, sz, fourier_pts, epsilon=1e-8, ntransforms=1, **kwargs):
        """
        A plan for non-uniform FFT in 2D or 3D.

        :param sz: A tuple indicating the geometry of the signal
        :param fourier_pts: The points in Fourier space where the Fourier transform is to be calculated,
            arranged as a dimension-by-K array. These need to be in the range [-pi, pi] in each dimension.
        :param epsilon: The desired precision of the NUFFT
        :param ntransforms: Optional integer indicating if you would like to compute a batch of `ntransforms`
        transforms.  Implies vol_f.shape is (..., `ntransforms`). Defaults to 0 which disables batching.
        """

        # Passing "ntransforms" > 1 expects one large higher dimensional array later.
        self.ntransforms = ntransforms

        # Workaround cufinufft A100 singles issue
        # ASPIRE-Python/703
        # Cast to doubles.
        self._original_dtype = fourier_pts.dtype
        fourier_pts = fourier_pts.astype(np.float64, copy=False)

        # Basic dtype passthough.
        dtype = fourier_pts.dtype
        if dtype == np.float64 or dtype == np.complex128:
            self.dtype = np.float64
            self.complex_dtype = np.complex128
        elif dtype == np.float32 or dtype == np.complex64:
            self.dtype = np.float32
            self.complex_dtype = np.complex64
        else:
            raise RuntimeError("Unsupported dtype encountered")

        self.sz = sz
        self.dim = len(sz)

        if not fourier_pts.flags.c_contiguous:
            logger.debug(
                "cufinufft has caught a non C_CONTIGUOUS array,"
                " `fourier_pts` will be copied to C_CONTIGUOUS."
            )
        self.fourier_pts = np.ascontiguousarray(
            np.mod(fourier_pts + np.pi, 2 * np.pi) - np.pi, dtype=self.dtype
        )

        self.num_pts = fourier_pts.shape[1]
        self.epsilon = max(epsilon, np.finfo(self.dtype).eps)

        self._transform_plan = cufinufft(
            2, self.sz, self.ntransforms, self.epsilon, -1, dtype=self.dtype
        )

        self.adjoint_opts = dict()
        if self.dtype is np.float64 and self.dim == 3 and self.epsilon < 1e3:
            # Note this is an algorithmic implementation dictated by shmem.
            logger.info(
                "Converting cufinufft gpu_method=1 from default of 2 for 3D1 transform,"
                f"to support computation in double precision with tol={self.epsilon}."
            )
            self.adjoint_opts["gpu_method"] = 1

        self._adjoint_plan = cufinufft(
            1,
            self.sz,
            self.ntransforms,
            self.epsilon,
            1,
            dtype=self.dtype,
            **self.adjoint_opts,
        )

        # Note, I store self.fourier_pts_gpu so the GPUArrray life
        #   is tied to instance, instead of this method.
        self.fourier_pts_gpu = gpuarray.to_gpu(self.fourier_pts)

        self._transform_plan.set_pts(*self.fourier_pts_gpu)
        self._adjoint_plan.set_pts(*self.fourier_pts_gpu)

    def transform(self, signal):
        """
        Compute the NUFFT transform using this plan instance.

        :param signal: Signal to be transformed. For a single transform,
        this should be a a 1, 2, or 3D array matching the plan `sz`.
        For a batch, signal should have shape `(*sz, ntransforms)`.

        :returns: Transformed signal of shape `num_pts` or
        `(ntransforms, num_pts)`.
        """

        # Check we're not forcing a dtype workaround for ASPIRE-Python/703,
        #   then check if we have a dtype mismatch.
        # This avoids false positive complaint for the workaround.
        if (self._original_dtype == self.dtype) and not (
            signal.dtype == self.dtype or signal.dtype == self.complex_dtype
        ):
            logger.warning(
                "Incorrect dtypes passed to (a)nufft."
                " In the future this will be an error."
            )

        sig_shape = signal.shape
        res_shape = self.num_pts
        # Note, there is a corner case for ntransforms == 1.
        if self.ntransforms > 1 or (
            self.ntransforms == 1 and len(signal.shape) == self.dim + 1
        ):
            assert (
                len(signal.shape) == self.dim + 1
            ), f"For multiple transforms, {self.dim}D signal should be a {self.ntransforms} element stack of {self.sz}."

            assert (
                signal.shape[0] == self.ntransforms
            ), "For multiple transforms, signal stack length should match ntransforms {self.ntransforms}."

            sig_shape = signal.shape[1:]  # order...
            res_shape = (self.ntransforms, self.num_pts)

        assert (
            sig_shape == self.sz
        ), f"Signal frame to be transformed must have shape {self.sz}"

        signal_gpu = gpuarray.to_gpu(
            np.ascontiguousarray(signal, dtype=self.complex_dtype)
        )

        result_gpu = gpuarray.GPUArray(res_shape, dtype=self.complex_dtype)

        self._transform_plan.execute(result_gpu, signal_gpu)

        result = result_gpu.get()
        # ASPIRE-Python/703
        result = result.astype(complex_type(self._original_dtype), copy=False)

        return result

    def adjoint(self, signal):
        """
        Compute the NUFFT adjoint using this plan instance.

        :param signal: Signal to be transformed. For a single transform,
        this should be a a 1D array of len `num_pts`.
        For a batch, signal should have shape `(ntransforms, num_pts)`.

        :returns: Transformed signal `(sz)` or `(sz, ntransforms)`.
        """

        # Check we're not forcing a dtype workaround for ASPIRE-Python/703,
        #   then check if we have a dtype mismatch.
        # This avoids false positive complaint for the workaround.
        if (self._original_dtype == self.dtype) and not (
            signal.dtype == self.complex_dtype or signal.dtype == self.dtype
        ):
            logger.warning(
                "Incorrect dtypes passed to (a)nufft."
                " In the future this will be an error."
            )

        res_shape = self.sz
        # Note, there is a corner case for ntransforms == 1.
        if self.ntransforms > 1 or (self.ntransforms == 1 and len(signal.shape) == 2):
            assert (
                len(signal.shape) == 2
            ), f"For multiple {self.dim}D adjoints, signal should be a {self.ntransforms} element stack of {self.num_pts}."
            assert (
                signal.shape[0] == self.ntransforms
            ), "For multiple transforms, signal stack length should match ntransforms {self.ntransforms}."
            res_shape = (self.ntransforms, *self.sz)

        signal_gpu = gpuarray.to_gpu(
            np.ascontiguousarray(signal, dtype=self.complex_dtype)
        )

        result_gpu = gpuarray.GPUArray(res_shape, dtype=self.complex_dtype)

        self._adjoint_plan.execute(signal_gpu, result_gpu)

        result = result_gpu.get()
        # ASPIRE-Python/703
        result = result.astype(complex_type(self._original_dtype), copy=False)

        return result



from .types import complex_type, real_type, utest_tolerance  # isort:skip
from .coor_trans import (  # isort:skip
    common_line_from_rots,
    crop_pad_2d,
    crop_pad_3d,
    get_aligned_rotations,
    get_rots_mse,
    grid_1d,
    grid_2d,
    grid_3d,
    register_rotations,
    rots_to_clmatrix,
    uniform_random_angles,
)

from .misc import (  # isort:skip
    all_pairs,
    all_triplets,
    abs2,
    bump_3d,
    circ,
    cyclic_rotations,
    gaussian_1d,
    gaussian_2d,
    gaussian_3d,
    importlib_path,
    inverse_r,
    J_conjugate,
    powerset,
    sha256sum,
    support_mask,
    fuzzy_mask,
)

from .logging import LogFilterByCount, get_full_version, tqdm, trange
from .matrix import (
    acorr,
    ainner,
    anorm,
    best_rank1_approximation,
    eigs,
    fix_signs,
    im_to_vec,
    make_psd,
    make_symmat,
    mat_to_vec,
    mdim_mat_fun_conj,
    roll_dim,
    symmat_to_vec,
    symmat_to_vec_iso,
    unroll_dim,
    vec_to_im,
    vec_to_mat,
    vec_to_symmat,
    vec_to_symmat_iso,
    vec_to_vol,
    vecmat_to_volmat,
    vol_to_vec,
    volmat_to_vecmat,
)
from .multiprocessing import (
    mem_based_cpu_suggestion,
    num_procs_suggestion,
    physical_core_cpu_suggestion,
    virtual_core_cpu_suggestion,
)
from .random import Random, choice, rand, randi, randn, random
from .relion_interop import RelionStarFile, relion_metadata_fields
from .resolution_estimation import FourierRingCorrelation, FourierShellCorrelation
from .rotation import Rotation
from .units import ratio_to_decibel, voltage_to_wavelength, wavelength_to_voltage

from .bot_align import align_BO  # isort:skip



import numpy as np
from scipy.linalg import eigh_tridiagonal


class BNMatrix:
    """
    Define a class to compute the B_N matrix ( with elements of b^N_mn) denoting the matrix of the operator L_c
    with respect to the basis T_Nn(x) by elements of b^N_mn. The matrix element b^N_mn is calculated as
    < T_Nm, L_c T_Nn > as shown in the paper below:
      Yoel Shkolnisky, "Prolate spheroidal wave functions on a disc-Integration and approximation of
      two-dimensional bandlimited functions", Appl. Comput. Harmon. Anal. 22, 235-256 (2007).
    """

    def __init__(self, big_n, band_limit, approx_length, dtype=np.float32):
        """
        Initial an object to compute the B_N matrix ( with elements of b^N_mn).

        :param big_n: A positive integer represented by N.
        :param band_limit: The band limit.
        :param approx_length: The approximated length represented by n.
        """

        self.dtype = np.dtype(dtype)

        k = np.arange(1, approx_length, dtype=float)
        self.diagonal = np.ones(approx_length, dtype=self.dtype)
        self.diagonal[0] = self._generate_bn_mat_b_n_on_diagonal(big_n, 0, band_limit)
        # BN matrix is a symmetric tridiagonal matrix
        self.diagonal[1:] = self._generate_bn_mat_b_n_on_diagonal(big_n, k, band_limit)
        self.off_diagonal = self._generate_bn_mat_b_n_above_diagonal(
            big_n, k, band_limit
        )
        self.off_diagonal += self._generate_bn_mat_b_n_below_diagonal(
            big_n, k - 1, band_limit
        )
        self.off_diagonal /= 2

    def get_eig_vectors(self):
        """
        Calculated the eigenvalues and eigenvectors of B_N matrix.

        :return: v: (M,M) ndarray
                    The normalized eigenvectors corresponding to the eigenvalues, v[:, i] is corresponding to the w[i].
                    In each eigenvector v[:, i], v[argmax(abs(v[:, i])), i] >= 0.
                w: (M,) ndarray
                    The eigenvalues in descending order.
        """

        w, v = eigh_tridiagonal(self.diagonal, self.off_diagonal, select="a")

        sorted_idx = np.argsort(-w)
        v = v[:, sorted_idx]
        w = w[sorted_idx]
        # We need to rescale the eigenvectors to fix the sign problem and make consistent with
        # the Matlab version.
        a = np.argmax(np.absolute(v), axis=0)
        b = np.array([np.sign(v[a[k], k]) for k in range(len(v))], dtype=self.dtype)
        v = v * b

        return v, w

    def dense_mat(self):
        """
        Represent the full B_N matrix by a 2D array.

        :return: mat: (M,M), ndarray
                The full BN matrix.
        """

        diagonal = self.diagonal
        off_diagonal = self.off_diagonal
        m = len(diagonal)

        mat = np.zeros((m, m), dtype=self.dtype)
        mat[0, 0] = diagonal[0]

        for i in range(m - 1):
            mat[i, i + 1] = off_diagonal[i]
            mat[i + 1, i] = off_diagonal[i]
            mat[i + 1, i + 1] = diagonal[i + 1]

        return mat

    def shape(self):
        """
        Ruturn the shape of B_N matrix.

        :return: tuple: (n, n)
            The dense matrix shape
        """

        return len(self.diagonal), len(self.diagonal)

    def _generate_bn_mat_h(self, n, k):
        """
        Defined in the paper eq (17) (basic equation) + (19) (the usage equation)
        """

        return np.sqrt(2 * (2 * k + n + 1))

    def _generate_bn_mat_k(self, n, k):
        """
        Defined in the paper below eq (20)
        """

        return (n + 2 * k + 0.5) * (n + 2 * k + 1.5)

    def _generate_bn_mat_gamma_plus_1(self, n, k):
        """
        Defined in the paper eq (24)
        """

        return -(
            (np.square(k + n + 1) * self._generate_bn_mat_h(n, k))
            / ((2 * k + n + 1) * (2 * k + n + 2) * self._generate_bn_mat_h(n, k + 1))
        ) * ((k + 1) / (k + n + 1))

    def _generate_bn_mat_gamma_0(self, n, k):
        """
        Defined in the paper eq (24)
        """

        if n == 0:
            return 0.5
        return (2.0 * k * (k + 1) + n * (2 * k + n + 1)) / (
            (2 * k + n) * (2 * k + n + 2)
        )

    def _generate_bn_mat_gamma_minus_1(self, n, k):
        """
        Defined in the paper eq (24)
        """

        return -(
            (np.square(k) * self._generate_bn_mat_h(n, k))
            / ((2 * k + n + 1) * (2 * k + n) * self._generate_bn_mat_h(n, k - 1))
        ) * ((n + k) / k)

    def _generate_bn_mat_b_n_above_diagonal(self, n, k, c):
        """
        Defined in the paper eq (26)
        """

        return -np.square(c) * self._generate_bn_mat_gamma_minus_1(n, k)

    def _generate_bn_mat_b_n_on_diagonal(self, n, k, c):
        """
        Defined in the paper eq (26)
        """

        return -(
            self._generate_bn_mat_k(n, k)
            + np.square(c) * self._generate_bn_mat_gamma_0(n, k)
        )

    def _generate_bn_mat_b_n_below_diagonal(self, n, k, c):
        """
        Defined in the paper eq (26)
        """

        return -np.square(c) * self._generate_bn_mat_gamma_plus_1(n, k)



import logging

import numpy as np
from scipy.linalg import eigh

from aspire.abinitio import CLSymmetryC3C4
from aspire.utils import J_conjugate, Rotation, all_pairs

logger = logging.getLogger(__name__)


class CLSymmetryC2(CLSymmetryC3C4):
    """
    Define a class to estimate 3D orientations using common lines methods for molecules with C2 cyclic symmetry.

    The related publications are listed below:
    X. Cheng,
    Random Matrices in High-dimensional Data Analysis,
    PhD thesis, Princeton University, (2013).

    G. Pragier and Y. Shkolnisky,
    A Common Lines Approach for Abinitio Modeling of Cyclically Symmetric Molecules,
    Inverse Problems, 35, 124005, (2019).

    Y. Shkolnisky, and A. Singer,
    Viewing Direction Estimation in Cryo-EM Using Synchronization,
    SIAM J. Imaging Sciences, 5, 1088-1110 (2012).

    A. Singer, R. R. Coifman, F. J. Sigworth, D. W. Chester, Y. Shkolnisky,
    Detecting Consistent Common Lines in Cryo-EM by Voting,
    Journal of Structural Biology, 169, 312-322 (2010).
    """

    def __init__(
        self,
        src,
        n_rad=None,
        n_theta=None,
        max_shift=0.15,
        shift_step=1,
        epsilon=1e-3,
        max_iters=1000,
        degree_res=1,
        min_dist_cls=25,
        seed=None,
    ):
        """
        Initialize object for estimating 3D orientations for molecules with C2 symmetry.

        :param src: The source object of 2D denoised or class-averaged images with metadata
        :param n_rad: The number of points in the radial direction
        :param n_theta: The number of points in the theta direction
        :param max_shift: Maximum range for shifts as a proportion of resolution. Default = 0.15.
        :param shift_step: Resolution of shift estimation in pixels. Default = 1 pixel.
        :param epsilon: Tolerance for the power method.
        :param max_iter: Maximum iterations for the power method.
        :param degree_res: Degree resolution for estimating in-plane rotations.
        :param min_dist_cls: Minimum distance between mutual common-lines. Default = 25 degrees.
        :param seed: Optional seed for RNG.
        """
        super().__init__(
            src,
            symmetry="C2",
            n_rad=n_rad,
            n_theta=n_theta,
            max_shift=max_shift,
            shift_step=shift_step,
            epsilon=epsilon,
            max_iters=max_iters,
            degree_res=degree_res,
            seed=seed,
        )

        self.min_dist_cls = min_dist_cls
        self.epsilon = epsilon
        self.max_iters = max_iters
        self.degree_res = degree_res
        self.seed = seed
        self.order = 2

    def _check_symmetry(self, symmetry):
        symmetry = symmetry.upper()
        if symmetry != "C2":
            raise NotImplementedError(
                f"Only C2 symmetry supported. {symmetry} was supplied."
            )

    def build_clmatrix(self):
        """
        Build common-lines matrix for molecules with C2 symmetry from Fourier stack of 2D images.
        This consists of finding for each pair of images the two common-lines induced by the 2-fold symmetry.
        """

        n_img = self.n_img

        if self.n_theta % 2 == 1:
            msg = "n_theta must be even"
            logger.error(msg)
            raise NotImplementedError(msg)

        # need to do a copy to prevent modifying self.pf for other functions.
        pf = self.pf.copy()

        # clmatrix contains the index in image i of the common line with image j
        # for the two sets of mutual common lines.
        clmatrix = -np.ones((2, n_img, n_img), dtype=int)

        # 1D shift between common-lines.
        shifts_1d = np.zeros((2, n_img, n_img))

        # Prepare the shift phases and generate filter for common-line detection.
        r_max = pf.shape[2]
        shifts, shift_phases, h = self._generate_shift_phase_and_filter(
            r_max, self.max_shift, self.shift_step
        )
        n_shifts = len(shifts)

        # Apply bandpass filter, normalize each ray of each image.
        # Note that only use half of each ray.
        pf = self._apply_filter_and_norm("ijk, k -> ijk", pf, r_max, h)

        # Pre-compute conjugated and shifted pf's.
        pf_shifted_flipped = np.conj(pf)[:, None] * shift_phases[:, None]
        pf_shifted_flipped = pf_shifted_flipped.reshape(
            (self.n_img, n_shifts * (self.n_theta // 2), r_max)
        )

        # Search for common lines between [i, j] pairs of images.
        for i in range(n_img - 1):
            p1 = pf[i]
            for j in range(i + 1, n_img):
                p2 = pf_shifted_flipped[j]
                corr = self._compute_correlations(p1, p2)
                corr = corr.reshape(self.n_theta, n_shifts, self.n_theta // 2)

                # Find first set of common-lines between the pair of images.
                first_cl1, first_shift, first_cl2 = np.unravel_index(
                    corr.argmax(), corr.shape
                )
                clmatrix[0, i, j] = first_cl1
                clmatrix[0, j, i] = first_cl2
                shifts_1d[0, i, j] = shifts[first_shift]

                # Mask corr around first set of common-lines to search for
                # second set of mutual common-lines.
                mask_dist = self.min_dist_cls * 2 * self.n_theta // 360
                corr_masked = corr * self._square_mask(
                    corr, first_cl1, first_cl2, mask_dist
                )

                # Find second set of mutual common-lines.
                second_cl1, second_shift, second_cl2 = np.unravel_index(
                    corr_masked.argmax(), corr.shape
                )
                clmatrix[1, i, j] = second_cl1
                clmatrix[1, j, i] = second_cl2
                shifts_1d[1, i, j] = shifts[second_shift]

        self.clmatrix = clmatrix
        self.shifts_1d = shifts_1d

    @staticmethod
    def _compute_correlations(a, b):
        """
        Compute the correlation between all pairs of lines in the polar Fourier images a and b.
        It is assumed that the polar Fourier images a and b include only half of the full set of
        rays, ie. in the range [0, 180) degrees. The correlation is expanded by computing the
        correlation with b and conj(b) to include the full set of rays associated with image b.

        :param a: 2D array size n_theta_a x radial_points.
        :param b: 2D array size n_theta_b x radial_points.

        :return: Correlation array of size n_theta_a x n_theta_b.
        """

        # Compute corrrelations in the positive and negative r directions.
        corr = np.zeros((2 * len(a), len(b)), dtype=a.dtype)
        corr[0 : len(a)] = np.real(a @ b.T)
        corr[len(a) :] = np.real(a @ np.conj(b).T)

        return corr

    @staticmethod
    def _square_mask(corr, theta_1, theta_2, dist):
        """
        For each shift we mask the correlation around the point (theta_1, theta_2)
        with a square mask of half-length `dist`.

        :param corr: Correlation array of shape (n_theta, n_shifts, n_theta // 2)
        :param theta_1: theta_1-coordinate for center of mask.
        :param theta_2: theta_2-coordinate for center of mask.
        :param dist: The distance from center to mask off.

        :return: Mask with square hole centered at (theta_1, theta_2) with shape = corr.shape.
        """
        n_theta, n_shifts, n_theta_half = corr.shape

        # Build mask.
        left = max(0, theta_1 - dist)
        right = min(n_theta, theta_1 + dist)
        bottom = max(0, theta_2 - dist)
        top = min(n_theta_half, theta_2 + dist)
        mask = np.ones((n_theta, n_theta_half), dtype=int)
        mask[left:right, bottom:top] = 0

        # Expand along shift axis.
        mask = np.repeat(mask, n_shifts, axis=0).reshape(corr.shape)

        return mask

    def estimate_rotations(self):
        """
        Estimate rotation matrices for molecules with C2 symmetry.
        """
        vijs, Rijs, Rijgs = self._estimate_relative_viewing_directions()

        logger.info("Estimating third rows of rotation matrices.")
        # The diagonal blocks of the 3n x 3n block matrix of relative rotations are
        # the identity since vi.T @ vi = I.
        viis = np.vstack((np.eye(3, dtype=self.dtype),) * self.n_img).reshape(
            self.n_img, 3, 3
        )
        vis = self._estimate_third_rows(vijs, viis)

        logger.info("Estimating in-plane rotations and rotations matrices.")
        Ris = self._estimate_inplane_rotations(vis, Rijs, Rijgs)

        self.rotations = Ris

    ###########################################
    # Primary Methods                         #
    ###########################################

    def _estimate_relative_viewing_directions(self):
        """
        Estimate the relative viewing directions vij = vi*vj^T, i<j, and vii = vi*vi^T, where
        vi is the third row of the i'th rotation matrix Ri.
        """
        logger.info(f"Estimating relative viewing directions for {self.n_img} images.")
        # Step 1: Detect the two pairs of mutual common-lines between each pair of images
        self.build_clmatrix()

        # Step 2: Calculate relative rotations associated with both mutual common lines.
        Rijs, Rijgs = self._estimate_all_Rijs_c2(self.clmatrix)

        # Step 3: Inner J-synchronization
        Rijs, Rijgs = self._local_J_sync_c2(Rijs, Rijgs)

        # Step 4: Global J-synchronization.
        logger.info("Performing global handedness synchronization.")
        vijs, Rijs, Rijgs = self._global_J_sync(Rijs, Rijgs)

        return vijs, Rijs, Rijgs

    def _global_J_sync(self, Rijs, Rijgs):
        """
        Global handedness synchronization of relative rotations, Rijs and Rijgs,
        and relative viewing directions vijs.

        :param Rijs: Relative rotations between pairs of images, shape n_pairs x 3 x 3.
        :param Rijgs: Second set of relative rotations between pairs of images, shape n_pairs 3 x 3.
        :returns:
            - vijs - Globally synchronized relative viewing directions.
            - Rijs - Globally synchronized relative rotations.
            - Rijgs - Globally synchronized 2nd set of relative rotations.
        """
        # Here we use the fact that the mean over the set of relative rotations induced
        # by the cyclic symmetry gives us the outer product of the third rows of the rotations
        # from the i'th and j'th images. See corollary 6.2 (G. Pragier) for more details.
        vijs = (Rijs + Rijgs) / 2

        # Determine relative handedness of vijs.
        sign_ij_J = self._J_sync_power_method(vijs)

        # Synchronize relative rotations
        for i, sign in enumerate(sign_ij_J):
            if sign == -1:
                vijs[i] = J_conjugate(vijs[i])
                Rijs[i] = J_conjugate(Rijs[i])
                Rijgs[i] = J_conjugate(Rijgs[i])

        return vijs, Rijs, Rijgs

    def _estimate_inplane_rotations(self, vis, Rijs, Rijgs):
        """
        Estimate rotation matrices for each image by first constructing arbitrary rotations with
        the given third rows, vis, then applying in-plane rotations found with an angular
        synchronization procedure.

        :param vis: Estimated third rows of orientation matrices, shape n_img x 3.
        :param Rijs: First set of J-synchronized relative rotations, shape n_pairs x 3 x 3.
        :param Rijgs: Second set of J-synchronized relative rotations, shape n_pairs x 3 x 3.
        :return: Estimated rotations, Ris, shape n_img x 3 x 3.
        """
        H = np.zeros((self.n_img, self.n_img), dtype=complex)
        # Step 1: Construct all rotation matrices Ris_tilde whose third rows are equal to
        # the corresponding third rows vis.
        Ris_tilde = self._complete_third_row_to_rot(vis)

        pairs = all_pairs(self.n_img)
        for idx, (i, j) in enumerate(pairs):
            # Uij and Uijg below are xy-in-plane rotations.
            # We use svd to find the in-plane rotation which is closest to
            # Ris_tilde[i] @ Rijs[idx] @ Ris_tilde[j].T.
            # See section 8.3.1 (X. Cheng) for more details.
            Uij = Ris_tilde[i] @ Rijs[idx] @ Ris_tilde[j].T
            u, _, v = np.linalg.svd(Uij[0:2, 0:2])
            Uij = u @ v.T

            Uijg = Ris_tilde[i] @ Rijgs[idx] @ Ris_tilde[j].T
            u, _, v = np.linalg.svd(Uijg[0:2, 0:2])
            Uijg = u @ v.T

            U = (Uij @ Uij + Uijg @ Uijg) / 2
            u, _, v = np.linalg.svd(U)
            U = u @ v.T

            H[i, j] = U[0, 0] - 1j * U[1, 0]

        # Populate the lower triangle and diagonal of H.
        # Diagonals are 1 since e^{i*0}=1.
        H += np.conj(H).T + np.eye(self.n_img)

        # H is a rank-1 Hermitian matrix.
        eig_vals, eig_vecs = eigh(H, subset_by_index=[self.n_img - 5, self.n_img - 1])
        leading_eig_vec = eig_vecs[:, -1]
        logger.info(f"Top 5 eigenvalues of H are {str(eig_vals[::-1])}.")

        # Calculate R_thetas.
        R_thetas = Rotation.about_axis("z", np.angle(np.sqrt(leading_eig_vec)))

        # Form rotation matrices Ris.
        Ris = R_thetas @ Ris_tilde

        return Ris

    #################################################
    # Secondary Methods for computing outer product #
    #################################################

    def _estimate_all_Rijs_c2(self, clmatrix):
        """
        Estimate the two sets of relative rotations, Rijs and Rijgs, between pairs
        of images using the voting method.

        :param clmatrix: 2 x n_img x n_img array holding two sets of mutual common-lines
            between pairs of images.
        :return: Relative rotations, Rijs and Rijgs.
        """
        k_list = np.arange(self.n_img)
        n_theta = self.n_theta
        pairs = all_pairs(self.n_img)
        Rijs = np.zeros((len(pairs), 3, 3), dtype=self.dtype)
        Rijgs = np.zeros((len(pairs), 3, 3), dtype=self.dtype)
        for idx, (i, j) in enumerate(pairs):
            Rijs[idx] = self._syncmatrix_ij_vote_3n(clmatrix[0], i, j, k_list, n_theta)
            Rijgs[idx] = self._syncmatrix_ij_vote_3n(clmatrix[1], i, j, k_list, n_theta)

        return Rijs, Rijgs

    def _local_J_sync_c2(self, Rijs, Rijgs):
        """
        For each pair of images J-synchronize the two corresponding relative rotations.

        :param Rijs: First set of relative rotations.
        :param Rijgs: Second set of relative rotations.

        Return: Pairwise J-synchronized relative rotations, Rijs and Rijgs.
        """
        e1 = np.array([1, 0, 0], dtype=self.dtype)
        vijs = (Rijs + Rijgs) / 2
        vijs_J = (J_conjugate(Rijs) + Rijgs) / 2

        s = np.linalg.svd(vijs, compute_uv=False)
        s_J = np.linalg.svd(vijs_J, compute_uv=False)

        # Indices to conjugate.
        ind_to_conj = np.argwhere(
            np.linalg.norm(s_J - e1, axis=1) < np.linalg.norm(s - e1, axis=1)
        )
        Rijgs[ind_to_conj] = J_conjugate(Rijgs[ind_to_conj])

        return Rijs, Rijgs



"""
Functions for compatibility with MATLAB behavior.
At some point when the package is full validated against MatLab, the 'order' arguments in the functions here
can be changed to 'C', and subsequently, this package deprecated altogether (i.e. the reshape/flatten methods used
directly by the caller).
"""

import numpy as np
import scipy.sparse as sparse


def m_reshape(x, new_shape):
    # This is a somewhat round-about way of saying:
    #   return x.reshape(new_shape, order='F')
    # We follow this approach since numba/cupy don't support the 'order'
    # argument, and we may want to use those decorators in the future
    # Note that flattening is required before reshaping, because
    if isinstance(new_shape, tuple):
        return m_flatten(x).reshape(new_shape[::-1]).T
    else:
        return x


def m_flatten(x):
    # This is a somewhat round-about way of saying:
    #   return x.flatten(order='F')
    # We follow this approach since numba/cupy don't support the 'order'
    # argument, and we may want to use those decorators in the future
    return x.T.flatten()


def stable_eigsh(*args, **kwargs):
    """
    A Wrapper function to fix sign problem of eigen-vectors

    There is an ambiguous sign problem for the eigenvectors from
    scipy.sparse.linalg.eigsh function. We need to rescale the
    eigenvectors and make them consistent for repeated runs.

    :param *args: Positional arguments
    :param **kwargs: Keyword arguments
    :return: Eigenvalues and eigenvectors
    """

    d, v = sparse.linalg.eigsh(*args, **kwargs)
    # Find component index of maximum absolute value of each eigenvector
    ind_max = np.argmax(np.absolute(v), axis=0)
    # Rescale eigenvector based on sign from the component with the
    # maximum absolute value
    signs = np.array([np.sign(v[ind_max[k], k]) for k in range(len(d))])

    return d, v * signs



import logging

import numpy as np

from aspire.noise import WhiteNoiseEstimator
from aspire.numeric import fft
from aspire.source import ImageSource
from aspire.utils import grid_2d

logger = logging.getLogger(__name__)


def adaptive_support(img_src, energy_threshold=0.99):
    """
    Determine size of the compact support in both real and Fourier Space.

    Returns c_limit (support radius in Fourier space),
    and R_limit (support radius in real space).

    Fourier c_limit is scaled in range [0, 0.5].
    R_limit is in pixels [0, Image.resolution/2].

    :param img_src: Input `Source` of images.
    :param energy_threshold: [0, 1] threshold limit
    :return: (c_limit, R_limit)
    """

    if not isinstance(img_src, ImageSource):
        raise RuntimeError("adaptive_support expects `Source` instance or subclass.")

    # Sanity Check Threshold is in range
    if energy_threshold <= 0 or energy_threshold > 1:
        raise ValueError(
            f"Given energy_threshold {energy_threshold} outside sane range [0,1]"
        )

    L = img_src.L
    N = L // 2

    r = grid_2d(L, shifted=False, normalized=False, indexing="yx", dtype=img_src.dtype)[
        "r"
    ]

    # Estimate noise
    noise_est = WhiteNoiseEstimator(img_src)
    noise_var = noise_est.estimate()

    # Transform to Fourier space
    img = img_src.images[:].asnumpy()
    imgf = fft.centered_fft2(img)

    # Compute the Variance and Power Spectrum
    #   Mean along image stack.
    variance_map = np.mean(np.abs(img) ** 2, axis=0)
    pspec = np.mean(np.abs(imgf) ** 2, axis=0)

    # Compute the Radial Variance and Radial Power Spectrum
    radial_var = np.zeros(N)
    radial_pspec = np.zeros(N)
    for i in range(N):
        mask = (r >= i) & (r < i + 1)
        # Mean along radial track defined by mask
        radial_var[i] = np.mean(variance_map[mask])
        radial_pspec[i] = np.mean(pspec[mask])

    # Subtract the noise variance
    radial_pspec -= noise_var
    radial_var -= noise_var

    # Lower bound variance and power by 0
    np.clip(radial_pspec, 0, a_max=None, out=radial_pspec)
    np.clip(radial_var, 0, a_max=None, out=radial_var)

    # Construct range of Fourier limits. We need a half-sample correction
    # since each ring is centered between two integer radii. Same for spatial
    # domain (R).
    c = (np.arange(N) + 0.5) / (2 * N)
    R = np.arange(N) + 0.5

    # Calculate cumulative energy
    cum_pspec = np.cumsum(radial_pspec * c)
    cum_var = np.cumsum(radial_var * R)

    # Normalize energies [0,1]
    #  Multiply threshold to avoid unstable division
    c_energy_threshold = energy_threshold * cum_pspec[-1]
    R_energy_threshold = energy_threshold * cum_var[-1]

    # First note legacy code *=L for Fourier limit,
    #   but then only uses divided by L... so just removed here.
    #   This makes it consistent with Nyquist, ie [0, .5]
    # Second note, we attempt to find the cutoff,
    #   but when a search fails returns the last (-1) element,
    #   essentially the maximal radius.
    # Third note, to increase accuracy, we take a weighted average of the two
    #   points around the cutoff. This mostly affects c since R is rounded.

    ind = np.argmax(cum_pspec > c_energy_threshold)
    if ind > 0:
        c_limit = (cum_pspec[ind - 1] * c[ind - 1] + cum_pspec[ind] * c[ind]) / (
            cum_pspec[ind - 1] + cum_pspec[ind]
        )
    else:
        c_limit = c[-1]

    ind = np.argmax(cum_var > R_energy_threshold)
    if ind > 0:
        R_limit = round(
            (cum_var[ind - 1] * R[ind - 1] + cum_var[ind] * R[ind])
            / (cum_var[ind - 1] + cum_var[ind])
        )
    else:
        R_limit = R[-1]

    return c_limit, R_limit



"""
This module contains code for estimating resolution achieved by reconstructions.
"""
import logging

import matplotlib.pyplot as plt
import numpy as np

from aspire.nufft import nufft
from aspire.numeric import fft
from aspire.utils import grid_2d, grid_3d

logger = logging.getLogger(__name__)

# FourierCorrelation holds a single implementation for both FSC and
# FRC based on dimension `dim`.


class FourierCorrelation:
    r"""
    Compute the Fourier correlations between two arrays.

    Underlying data (images/volumes) are assumed to be well aligned.

    The Fourier correlation is defined as:

    .. math::

       c(i) = \frac{ \operatorname{Re}( \sum_i{ \mathcal{F}_1(i) * {\mathcal{F}^{*}_2(i) } } ) }{\
         \sqrt{ \sum_i { | \mathcal{F}_1(i) |^2 } * \sum_i{| \mathcal{F}^{*}_2}(i) |^2 } }

    This implementation supports Numpy style broadcasting resulting in
    up to two stack dimensions.  For example, to compute all pairs
    supply signal with stack shapes (m,1) and (1,n) to yield an (m,n)
    table of results. Note that plotting is limited to a single
    reference signal.
    """

    def __init__(self, a, b, pixel_size=None, method="fft"):
        """
                :param a: Input array a, shape(..., *dim).
                :param b: Input array b, shape(..., *dim).
                :param pixel_size: Pixel size in angstrom.
                    Default `None` implies "pixel" units.
                :param method: Selects either 'fft' (on Cartesian grid),
                    or 'nufft' (on polar grid). Defaults to 'fft'.
        7"""

        # Sanity checks
        if not hasattr(self, "dim"):
            raise RuntimeError("Subclass must assign `dim`")
        for x in (a, b):
            if not isinstance(x, np.ndarray):
                raise TypeError(f"`{x}` is not a Numpy array.")

        if not a.dtype == b.dtype:
            raise TypeError(
                f"Mismatched input types {a.dtype} != {b.dtype}. Cast `a` or `b`."
            )
        # TODO, check-math/avoid complex inputs.

        # Shape checks
        if not a.shape[-self.dim :] == b.shape[-self.dim :]:
            raise RuntimeError(
                f"`a` and `b` appear to have different data axis shapes, {a.shape} {b.shape}"
            )

        # Method selection
        methods = {"fft": self._fft_correlations, "nufft": self._nufft_correlations}
        if method not in methods:
            raise RuntimeError(
                f"Requested method {method} not in available methods {list(methods.keys())}."
            )
        self.method = method
        self._correlation_method = methods[self.method]

        # To support arbitrary broadcasting simply,
        # we'll force all shapes to be (-1, *(L,)*dim)
        # and keep track of the stack shapes.
        self.a = a
        self.b = b
        self._a, self._a_stack_shape = self._reshape(a)
        self._b, self._b_stack_shape = self._reshape(b)
        self._result_stack_shape = np.broadcast_shapes(
            self._a_stack_shape, self._b_stack_shape
        )

        # Handle `pixel_size` and `pixel_mode`
        self._pixel_units = "angstrom"
        if pixel_size is None:
            pixel_size = 1.0
            self._pixel_units = "pixels"
        self.pixel_size = float(pixel_size)

        self._correlations = None
        self.L = self._a.shape[-1]
        self.dtype = self._a.dtype

    @property
    def _fourier_axes(self):
        """
        Returns tuple representing the axes containing signal data
        based on dimension `dim`.
        """
        return tuple(range(-self.dim, 0))

    def _reshape(self, x):
        """
        Returns `x` with flattened stack axis and `x`'s original stack
        shape, as determined by `dim`.

        :param x: Numpy ndarray

        :return: (stack flattened x, x_stack_shape)
        """
        # TODO, check 2d input for dim=2 (singleton case)
        original_stack_shape = x.shape[: -self.dim]
        x = x.reshape(-1, *x.shape[-self.dim :])
        return x, original_stack_shape

    @property
    def correlations(self):
        """
        Compute and return the Fourier correlations of signal stacks a
        cross b.

        :return: Numpy array
        """
        # Cache _correlations.
        # There is no need to run this twice assuming inputs are immutable.
        if self._correlations is None:
            # Compute the correlations
            self._correlations = self._correlation_method()

        return self._correlations

    def _fft_correlations(self):
        """
        Computes Fourier correlations using the FFT on a Cartesian grid.
        """

        # Compute shells from 2D grid.
        if self.dim == 2:
            grid_function = grid_2d
        elif self.dim == 3:
            grid_function = grid_3d

        radii = grid_function(self.L, shifted=True, normalized=False, dtype=self.dtype)[
            "r"
        ]

        # Compute centered Fourier transforms.
        f1 = fft.centered_fftn(self.a, axes=self._fourier_axes)
        f2 = fft.centered_fftn(self.b, axes=self._fourier_axes)

        # Construct an output table of correlations
        correlations = np.zeros(
            (self.L // 2, *self._result_stack_shape), dtype=self.dtype
        )

        inner_diameter = 0.5
        for i in range(0, self.L // 2):
            # Compute ring mask
            outer_diameter = 0.5 + (i + 1)
            ring_mask = (radii > inner_diameter) & (radii < outer_diameter)
            logger.debug(f"Shell, Elements:  {i}, {np.sum(ring_mask)}")

            # Mask off values in Fourier space
            r1 = ring_mask * f1
            r2 = ring_mask * f2

            # Compute Fourier correlations
            num = np.real(np.sum(r1 * np.conj(r2), axis=self._fourier_axes))
            den = np.sqrt(
                np.sum(np.abs(r1) ** 2, axis=self._fourier_axes)
                * np.sum(np.abs(r2) ** 2, axis=self._fourier_axes)
            )
            # Assign
            correlations[i] = num / den
            # Update ring
            inner_diameter = outer_diameter

        # Repack the table as (..., L//2)
        correlations = np.swapaxes(correlations, 0, -1)
        return correlations.reshape(*self._result_stack_shape, self.L // 2)

    def _nufft_correlations(self):
        """
        Computes Fourier correlations using the NUFFT on a polar grid.
        """

        # TODO, we could use an internal tool (Polar2D?) for this.
        # L//2 is intentionally used for compatibility with Cartesian grid.
        #   This avoids having to have multiple methods for computing resolutions later.
        r = np.linspace(0, np.pi, self.L // 2, endpoint=False, dtype=self.dtype)
        phi = np.linspace(0, 2 * np.pi, 2 * self.L, endpoint=False, dtype=self.dtype)
        if self.dim == 2:
            # 2D Polar points
            x = r[:, np.newaxis] * np.cos(phi[np.newaxis, :])
            y = r[:, np.newaxis] * np.sin(phi[np.newaxis, :])
            # Because the values will be summed later, ordering does not matter.
            fourier_pts = np.vstack((x.flatten(), y.flatten()))

        elif self.dim == 3:
            # 3D Spherical points
            theta = np.linspace(0, np.pi, self.L, endpoint=False, dtype=self.dtype)
            x = (
                r[:, np.newaxis, np.newaxis]
                * np.sin(theta[np.newaxis, :, np.newaxis])
                * np.cos(phi[np.newaxis, np.newaxis, :])
            )
            y = (
                r[:, np.newaxis, np.newaxis]
                * np.sin(theta[np.newaxis, :, np.newaxis])
                * np.sin(phi[np.newaxis, np.newaxis, :])
            )
            z = (
                r[:, np.newaxis, np.newaxis]
                * np.cos(theta[np.newaxis, :, np.newaxis])
                * np.ones((1, 1, 2 * self.L), dtype=self.dtype)
            )
            # Because the values will be summed later, ordering does not matter.
            fourier_pts = np.vstack((x.flatten(), y.flatten(), z.flatten()))
        else:
            raise NotImplementedError(
                "`nufft` based correlations only implemented for dimensions 2 and 3."
            )

        # Stack signal data to create a larger NUFFT problem (better performance).
        #   Note, we want a complex result.
        signal = np.vstack((self._a, self._b))
        # Compute one large NUFFT for all the signal frames,
        f = nufft(signal, fourier_pts, real=False)
        # then unpack as two 1D stacks of the polar grid points, one for _a and _b.
        f = f.reshape(self._a.shape[0] + self._b.shape[0], len(r), -1)
        f1, f2 = np.vsplit(f, [self._a.shape[0]])

        # Compute the Fourier correlations.
        cov = np.sum(f1 * np.conj(f2), -1).real
        norm1 = np.sqrt(np.sum(np.abs(f1) ** 2, -1))
        norm2 = np.sqrt(np.sum(np.abs(f2) ** 2, -1))

        correlations = cov / (norm1 * norm2)

        # Then unpack as original a and b broadcasted shapes.
        return correlations.reshape(*self._result_stack_shape, r.shape[-1])

    def analyze_correlations(self, cutoff):
        """
        Convert from the Fourier correlations to frequencies and resolution.
        :param cutoff: Cutoff value, traditionally `.143`.
        """

        cutoff = float(cutoff)
        if not (0 <= cutoff <= 1):
            raise ValueError("Supplied correlation `cutoff` not in [0,1], {cutoff}")

        c_inds = np.zeros(self.correlations.shape[:-1], dtype=int)

        # All correlations are above cutoff,
        #   set index of highest sampled frequency.
        c_inds[np.min(self.correlations, axis=-1) > cutoff] = self.L // 2

        # Correlations cross the cutoff.
        # Find the first index of a correlation at `cutoff`.
        # Should return 0 if not found, which corresponded to the case
        # where all correlations are below cutoff.
        c_ind = np.maximum(c_inds, np.argmax(self.correlations <= cutoff, axis=-1))

        # Convert indices to frequency (as 1/angstrom)
        frequencies = self._freq(c_ind)

        # Convert to resolution in angstrom, smaller is higher frequency.
        self._resolutions = 1 / frequencies

        return self._resolutions

    def _freq(self, k):
        """
        Converts `k` from index of Fourier transform to frequency (as
        length 1/A).

        :param k: Frequency index, integer or Numpy array of ints.
        :return: Frequency in 1/angstrom.
        """

        # _freq(k) Units: 1 / (pixels * (angstrom / pixel) = 1 / angstrom
        # Similar to wavenumbers.  Larger is higher frequency.
        return k / (self.L * self.pixel_size)

    def plot(self, cutoff, save_to_file=False, labels=None):
        """
        Generates a Fourier correlation plot.

        :param cutoff: Cutoff value, traditionally `.143`.
        :param save_to_file: Optionally, save plot to file.
            Defaults False, enabled by providing a string filename.
            User is responsible for providing reasonable filename.
            See `https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.savefig.html`.
        """
        cutoff = float(cutoff)
        if not (0 <= cutoff <= 1):
            raise ValueError("Supplied correlation `cutoff` not in [0,1], {cutoff}")

        # Construct x-axis labels
        x_inds = np.arange(self.correlations.shape[-1])
        freqs = self._freq(x_inds)
        # TODO: handle zero freq better
        with np.errstate(divide="ignore"):
            freqs_units = 1 / freqs

        # Check we're asking for a reasonable plot.
        stack = self.correlations.shape[:-1]
        if len(stack) > 2:
            raise RuntimeError(
                f"Unable to plot figure tables with more than 2 dim, stack shape {stack}. Try reducing to a simpler request."
            )

        if (
            self._a_stack_shape[0] > 1 and self._a_stack_shape != self._b_stack_shape
        ) or (len(stack) == 2 and 1 not in stack):
            raise RuntimeError(
                f"Unable to plot figure tables with more than 1 reference figures, stack shape {stack}. Try reducing to a simpler request."
            )

        # Check `labels` length when provided.
        if labels is not None:
            if len(labels) != len(self.correlations):
                raise ValueError(
                    f"Check `labels`. Provided len(labels) != len(self.correlations): {len(labels)} != {len(self.correlations)}."
                )

        plt.figure(figsize=(8, 6))
        plt.title(self._plot_title)
        plt.xlabel(f"Resolution ({self._pixel_units})")
        plt.ylabel("Correlation")
        plt.ylim([0, 1.1])
        for i, line in enumerate(self.correlations):
            _label = None
            if len(self.correlations) > 1:
                _label = f"{i}"
                if labels is not None:
                    _label = labels[i]
            plt.plot(freqs_units, line, label=_label)

        # Display cutoff
        plt.axhline(y=cutoff, color="r", linestyle="--", label=f"cutoff={cutoff}")
        estimated_resolution = self.analyze_correlations(cutoff)[0]

        # Display resolution
        plt.axvline(
            x=estimated_resolution,
            color="b",
            linestyle=":",
            label=f"Resolution={estimated_resolution:.3f}",
        )
        # x-axis decreasing
        plt.gca().invert_xaxis()
        plt.legend(title=f"Method: {self.method}")

        if save_to_file:
            plt.savefig(save_to_file)
            plt.close()
        else:
            plt.show()


# The following are user facing classes, and simply wrap
# `FourierCorrelation` after assigning dimension `dim` and any
# dimension specific variables.


class FourierRingCorrelation(FourierCorrelation):
    """
    See `FourierCorrelation`.
    """

    dim = 2
    _plot_title = "Fourier Ring Correlation"


class FourierShellCorrelation(FourierCorrelation):
    """
    See `FourierCorrelation`.
    """

    dim = 3
    _plot_title = "Fourier Shell Correlation"



from .commonline_base import CLOrient3D
from .sync_voting import SyncVotingMixin

# isort: off
from .commonline_sync import CLSyncVoting
from .commonline_c3_c4 import CLSymmetryC3C4
from .commonline_cn import CLSymmetryCn
from .commonline_c2 import CLSymmetryC2

# isort: on



import logging

import numpy as np

from aspire.abinitio import CLOrient3D, SyncVotingMixin
from aspire.utils.matlab_compat import stable_eigsh

logger = logging.getLogger(__name__)


class CLSyncVoting(CLOrient3D, SyncVotingMixin):
    """
    Define a class to estimate 3D orientations using synchronization matrix and voting method.

    The related publications are listed as below:
    Y. Shkolnisky, and A. Singer,
    Viewing Direction Estimation in Cryo-EM Using Synchronization,
    SIAM J. Imaging Sciences, 5, 1088-1110 (2012).

    A. Singer, R. R. Coifman, F. J. Sigworth, D. W. Chester, Y. Shkolnisky,
    Detecting Consistent Common Lines in Cryo-EM by Voting,
    Journal of Structural Biology, 169, 312-322 (2010).
    """

    def __init__(self, src, n_rad=None, n_theta=360, max_shift=0.15, shift_step=1):
        """
        Initialize an object for estimating 3D orientations using synchronization matrix

        :param src: The source object of 2D denoised or class-averaged images with metadata
        :param n_rad: The number of points in the radial direction
        :param n_theta: The number of points in the theta direction.
            Default is 360.
        :param max_shift: Determines maximum range for shifts as a proportion
            of the resolution. Default is 0.15.
        :param shift_step: Resolution for shift estimation in pixels. Default is 1 pixel.
        """
        super().__init__(
            src,
            n_rad=n_rad,
            n_theta=n_theta,
            max_shift=max_shift,
            shift_step=shift_step,
        )
        self.syncmatrix = None

    def estimate_rotations(self):
        """
        Estimate orientation matrices for all 2D images using synchronization matrix
        """
        if self.syncmatrix is None:
            self.syncmatrix_vote()

        S = self.syncmatrix
        sz = S.shape
        assert sz[0] == sz[1], "syncmatrix must be a square matrix."
        assert sz[0] % 2 == 0, "syncmatrix must be a square matrix of size 2Kx2K."

        n_img = sz[0] // 2

        # S is a 2Kx2K matrix (K=n_img), containing KxK blocks of size 2x2.
        # The [i,j] block is given by [r11 r12; r12 r22], where
        # r_{kl}=<R_{i}^{k},R_{j}^{l}>, k,l=1,2, namely, the dot product of
        # column k of R_{i} and columns l of R_{j}. Thus, given the true
        # rotations R_{1},...,R_{K}, S is decomposed as S=W^{T}W where
        # W=(R_{1}^{1},R_{1}^{2},...,R_{K}^{1},R_{K}^{2}), where R_{j}^{k}
        # the k column of R_{j}. Therefore, S is a rank-3 matrix, and thus, it
        # three eigenvectors that correspond to non-zero eigenvalues, are linear
        # combinations of the column space of S, namely, W^{T}.

        # Extract three eigenvectors corresponding to non-zero eigenvalues.
        d, v = stable_eigsh(S, 10)
        sort_idx = np.argsort(-d)
        logger.info(
            f"Top 10 eigenvalues from synchronization voting matrix: {d[sort_idx]}"
        )

        # Only need the top 3 eigen-vectors.
        v = v[:, sort_idx[:3]]
        # According to the structure of W^{T} above, the odd rows of V, denoted V1,
        # are a linear combination of the vectors R_{i}^{1}, i=1,...,K, that is of
        # column 1 of all rotation matrices. Similarly, the even rows of V,
        # denoted, V2, are linear combinations of R_{i}^{1}, i=1,...,K.
        v1 = v[: 2 * n_img : 2].T.copy()
        v2 = v[1 : 2 * n_img : 2].T.copy()

        # We look for a linear transformation (3 x 3 matrix) A such that
        # A*V1'=R1 and A*V2=R2 are the columns of the rotations matrices.
        # Therefore:
        # V1 * A'*A V1' = 1
        # V2 * A'*A V2' = 1
        # V1 * A'*A V2' = 0
        # These are 3*K linear equations for 9 matrix entries of A'*A
        # Actually, there are only 6 unknown variables, because A'*A is symmetric.
        # So we will truncate from 9 variables to 6 variables corresponding
        # to the upper half of the matrix A'*A
        truncated_equations = np.zeros((3 * n_img, 6), dtype=self.dtype)
        k = 0
        for i in range(3):
            for j in range(i, 3):
                truncated_equations[0::3, k] = v1[i] * v1[j]
                truncated_equations[1::3, k] = v2[i] * v2[j]
                truncated_equations[2::3, k] = v1[i] * v2[j]
                k += 1

        # b = [1 1 0 1 1 0 ...]' is the right hand side vector
        b = np.ones(3 * n_img)
        b[2::3] = 0

        # Find the least squares approximation of A'*A in vector form
        ATA_vec = np.linalg.lstsq(truncated_equations, b, rcond=None)[0]

        # Construct the matrix A'*A from the vectorized matrix.
        ATA = np.zeros((3, 3), dtype=self.dtype)
        upper_mask = np.triu_indices(3)
        ATA[upper_mask] = ATA_vec
        lower_mask = np.tril_indices(3)
        ATA[lower_mask] = ATA.T[lower_mask]

        # The Cholesky decomposition of A'*A gives A
        # numpy returns lower, matlab upper
        a = np.linalg.cholesky(ATA)

        # Recover the rotations. The first two columns of all rotation
        # matrices are given by unmixing V1 and V2 using A. The third
        # column is the cross product of the first two.
        r1 = np.dot(a, v1)
        r2 = np.dot(a, v2)
        r3 = np.cross(r1, r2, axis=0)

        rotations = np.empty((n_img, 3, 3), dtype=self.dtype)
        rotations[:, :, 0] = r1.T
        rotations[:, :, 1] = r2.T
        rotations[:, :, 2] = r3.T
        # Make sure that we got rotations by enforcing R to be
        # a rotation (in case the error is large)
        u, _, v = np.linalg.svd(rotations)
        np.einsum("ijk, ikl -> ijl", u, v, out=rotations)

        self.rotations = rotations

    def syncmatrix_vote(self):
        """
        Construct the synchronization matrix using voting method

        A pre-computed common line matrix is required as input.
        """
        if self.clmatrix is None:
            self.build_clmatrix()

        clmatrix = self.clmatrix

        sz = clmatrix.shape
        n_theta = self.n_theta

        assert sz[0] == sz[1], "clmatrix must be a square matrix."

        n_img = sz[0]
        S = np.eye(2 * n_img, dtype=self.dtype).reshape(n_img, 2, n_img, 2)

        # Build Synchronization matrix from the rotation blocks in X and Y
        for i in range(n_img - 1):
            for j in range(i + 1, n_img):
                rot_block = self._syncmatrix_ij_vote(
                    clmatrix, i, j, np.arange(n_img), n_theta
                )
                S[i, :, j, :] = rot_block
                S[j, :, i, :] = rot_block.T

        self.syncmatrix = S.reshape(2 * n_img, 2 * n_img)

    def _syncmatrix_ij_vote(self, clmatrix, i, j, k_list, n_theta):
        """
        Compute the (i,j) rotation block of the synchronization matrix using voting method

        Given the common lines matrix `clmatrix`, a list of images specified in k_list
        and the number of common lines n_theta, find the (i, j) rotation block (in X and Y)
        of the synchronization matrix.
        :param clmatrix: The common lines matrix
        :param i: The i image
        :param j: The j image
        :param k_list: The list of images for the third image for voting algorithm
        :param n_theta: The number of points in the theta direction (common lines)
        :return: The (i,j) rotation block of the synchronization matrix
        """

        good_k = self._vote_ij(clmatrix, n_theta, i, j, k_list)

        rots = self._rotratio_eulerangle_vec(clmatrix, i, j, good_k, n_theta)

        if rots is not None:
            rot_mean = np.mean(rots, 0)
            # The error to mean value can be calculated as
            #    rot_block = rots[:2, :2]
            #    diff = rot_block - rot_mean[np.newaxis, :2, :2]
            #    err = np.linalg.norm(diff) / np.linalg.norm(rot_block)
            # if err > tol, this means that images i and j have inconsistent
            # rotations. The original Matlab code tried to print out the information
            # on inconsistent rotations but was commented out and do nothing,
            # probably due to the fact that it will print out a lot of
            # inconsistent rotations if the resolution or number of images
            # are not enough. We choose to pass it as Matlab code.

        else:
            # This for the case that images i and j correspond to the same
            # viewing direction and differ only by in-plane rotation.
            # Simply put to zero as Matlab code.
            rot_mean = np.zeros((3, 3))

        # return the rotation matrix in X and Y
        r22 = rot_mean[:2, :2]
        return r22



import numpy as np

from aspire.numeric import fft, xp


class PickerHelper:
    @classmethod
    def gaussian_filter(cls, size_filter, std):
        """Computes low-pass filter.

        Args:
            size_filter: Size of filter (size_filter x size_filter).
            std: sigma value in filter.
        """

        y, x = xp.mgrid[
            -(size_filter - 1) // 2 : (size_filter - 1) // 2 + 1,
            -(size_filter - 1) // 2 : (size_filter - 1) // 2 + 1,
        ]

        response = xp.exp(-xp.square(x) - xp.square(y) / (2 * (std**2))) / (
            xp.sqrt(2 * xp.pi) * std
        )
        response[response < xp.finfo("float").eps] = 0

        return xp.asnumpy(response / response.sum())  # Normalize so sum is 1

    @classmethod
    def extract_windows(cls, img, block_size):
        """Extracts blocks of size (block_size x block_size) from the micrograph. Blocks are
        extracted with steps of size (block_size)

        Args:
            img: Micrograph image.
            block_size: required block size.

        Returns:
            3D Matrix of blocks. For example, img[0] is the first block.
        """

        # Compute x,y boundary using block_size
        xbnd, ybnd = img.shape - np.mod(img.shape, block_size)
        # Keep only the portion of the image that can be split into blocks with no remainder
        # TODO: It might be nicer to apply any trimming to all micrograph sides
        img = xp.asarray(img[:xbnd, :ybnd])

        dim3_size = np.sqrt(np.prod(img.shape) // (block_size**2)).astype(int)

        img = xp.reshape(img, (block_size, dim3_size, block_size, dim3_size), "F")

        img = xp.transpose(img, (0, 2, 1, 3))
        img = xp.reshape(
            img, (img.shape[0] * img.shape[1], img.shape[2] * img.shape[3]), "F"
        )

        return img

    @classmethod
    def extract_query(cls, img, block_size):
        """Extract all query images from the micrograph. windows are
        extracted with steps of size (block_size/2)

        Args:
            img: Micrograph image.
            block_size: Query images must be of size (block_size x block_size).

        Returns:
            4D Matrix of query images.
        """

        # keep only the portion of the image that can be split into blocks with no remainder
        blocks = xp.asarray(
            img[
                : img.shape[0] - (img.shape[0] % block_size),
                : img.shape[1] - (img.shape[1] % block_size),
            ]
        )

        dim3_size = np.sqrt(np.prod(blocks.shape) // (block_size**2)).astype(int)
        blocks = xp.reshape(blocks, (block_size, dim3_size, block_size, dim3_size), "F")

        blocks = xp.transpose(blocks, (0, 2, 1, 3))

        blocks = xp.reshape(blocks, (blocks.shape[0], blocks.shape[1], -1), "F")

        blocks = xp.concatenate(
            (
                blocks,
                xp.concatenate(
                    (
                        blocks[:, :, 1:],
                        xp.reshape(
                            blocks[:, :, 0], (blocks.shape[0], blocks.shape[1], 1), "F"
                        ),
                    ),
                    axis=2,
                ),
            ),
            axis=0,
        )

        temp = xp.concatenate(
            (
                blocks[:, :, int(np.floor(2 * img.shape[1] / 2 / block_size)) :],
                blocks[:, :, 0 : int(np.floor(2 * img.shape[1] / 2 / block_size))],
            ),
            axis=2,
        )

        blocks = xp.concatenate((blocks, temp), axis=1)

        blocks = xp.reshape(
            blocks,
            (
                2 * block_size,
                2 * block_size,
                int(np.floor(2 * img.shape[0] / 2 / block_size)),
                int(np.floor(2 * img.shape[1] / 2 / block_size)),
            ),
            "F",
        )

        blocks = blocks[:, :, 0 : blocks.shape[2] - 1, 0 : blocks.shape[3] - 1]

        blocks = xp.transpose(blocks, (2, 3, 0, 1))

        return blocks

    @classmethod
    def reference_size(cls, img, container_size):
        num_containers_row = img.shape[0] // container_size
        num_containers_col = img.shape[1] // container_size

        return num_containers_row * num_containers_col * 4

    @classmethod
    def extract_references(cls, img, query_size, container_size):
        """Chooses and extracts reference images from the micrograph.

        Args:
            img: Micrograph image.
            query_size: Reference images must be of the same size of query images, i.e. (query_size x query_size).
            container_size: Containers are large regions used to select reference images. The size of each
                region is (container_size x container_size)

        Returns:
            3D Matrix of reference images.  windows[0] is the first reference window.
        """

        img = xp.asarray(img)
        num_containers_row = img.shape[0] // container_size
        num_containers_col = img.shape[1] // container_size

        windows = xp.zeros(
            (cls.reference_size(img, container_size), query_size, query_size)
        )
        win_idx = 0

        mean_all, std_all = cls.moments(img, query_size)

        for y_contain in range(num_containers_row):
            for x_contain in range(num_containers_col):
                temp = img[
                    y_contain
                    * container_size : min(
                        img.shape[0], (y_contain + 1) * container_size
                    ),
                    x_contain
                    * container_size : min(
                        img.shape[1], (x_contain + 1) * container_size
                    ),
                ]

                y_start = y_contain * container_size + query_size - 1
                y_end = min(
                    mean_all.shape[0] - query_size, (y_contain + 1) * container_size
                )

                x_start = x_contain * container_size + query_size - 1
                x_end = min(
                    mean_all.shape[1] - query_size, (x_contain + 1) * container_size
                )

                mean_contain = mean_all[y_start:y_end, x_start:x_end]
                std_contain = std_all[y_start:y_end, x_start:x_end]

                ind = xp.argmax(mean_contain)
                if ind.size == 1:
                    y, x = xp.unravel_index(ind, mean_contain.shape)
                    windows[win_idx, :, :] = temp[
                        y : y + query_size, x : x + query_size
                    ]
                    win_idx += 1

                ind = xp.argmin(mean_contain)
                if ind.size == 1:
                    y, x = xp.unravel_index(ind, mean_contain.shape)
                    windows[win_idx, :, :] = temp[
                        y : y + query_size, x : x + query_size
                    ]
                    win_idx += 1

                ind = xp.argmax(std_contain)
                if ind.size == 1:
                    y, x = xp.unravel_index(ind, std_contain.shape)
                    windows[win_idx, :, :] = temp[
                        y : y + query_size, x : x + query_size
                    ]
                    win_idx += 1

                ind = xp.argmin(std_contain)
                if ind.size == 1:
                    y, x = xp.unravel_index(ind, std_contain.shape)
                    windows[win_idx, :, :] = temp[
                        y : y + query_size, x : x + query_size
                    ]
                    win_idx += 1

        return windows

    @classmethod
    def get_training_set(cls, micro_img, bw_mask_p, bw_mask_n, n):
        """Gets training set for the SVM classifier.

        Args:
            micro_img: Micrograph image.
            bw_mask_p: Binary image indicating regions from which to extract examples of particles.
            bw_mask_n: Binary image indicating regions from which to extract examples of noise.
            n: Size of training windows.

        Returns:
            A matrix of features and a vector of labels for the SVM training.
        """

        non_overlap = cls.extract_windows(micro_img, n)

        indicate = cls.extract_windows(bw_mask_p, n)
        # consider only windows with no noise
        windows = non_overlap[:, (indicate != 0).all(axis=0)]
        p_mu = xp.mean(windows, axis=0)
        p_std = xp.std(windows, axis=0)

        indicate = cls.extract_windows(bw_mask_n, n)
        # consider only windows with no particles
        windows = non_overlap[:, (indicate != 1).all(axis=0)]
        n_mu = xp.mean(windows, axis=0)
        n_std = xp.std(windows, axis=0)

        p_mu = xp.reshape(p_mu, (p_mu.shape[0], 1), "F")
        p_std = xp.reshape(p_std, (p_std.shape[0], 1), "F")
        n_mu = xp.reshape(n_mu, (n_mu.shape[0], 1), "F")
        n_std = xp.reshape(n_std, (n_std.shape[0], 1), "F")

        x = xp.concatenate((p_mu, p_std), axis=1)
        x = xp.concatenate((x, xp.concatenate((n_mu, n_std), axis=1)), axis=0)

        y = xp.concatenate((xp.ones(p_mu.shape[0]), xp.zeros(n_mu.shape[0])), axis=0)

        return x, y

    @classmethod
    def moments(cls, img, query_size):
        """Calculates the mean and standard deviation for each window of size (query_size x query_size)
        in the micrograph.

        Args:
            img: Micrograph image.
            query_size: Size of windows for which to compute mean and std.

        Returns:
            A matrix of mean intensity and a matrix of variance, each containing a single
            entry for each possible (query_size x query_size) window in the micrograph.
        """

        filt = xp.ones((query_size, query_size)) / (query_size * query_size)
        filt = xp.pad(filt, (0, img.shape[0] - 1), "constant", constant_values=(0, 0))
        filt_freq = fft.fft2(filt, axes=(0, 1))

        pad_img = xp.pad(img, (0, query_size - 1), "constant", constant_values=(0, 0))
        img_freq = fft.fft2(pad_img, axes=(0, 1))

        mean_freq = xp.multiply(img_freq, filt_freq)
        mean_all = fft.ifft2(mean_freq, axes=(0, 1)).real

        pad_img_square = np.square(pad_img)
        img_var_freq = fft.fft2(pad_img_square, axes=(0, 1))
        var_freq = xp.multiply(img_var_freq, filt_freq)
        var_all = fft.ifft2(var_freq, axes=(0, 1))
        var_all = var_all.real - xp.power(mean_all, 2)
        std_all = xp.sqrt(xp.maximum(0, var_all))

        return mean_all, std_all



import logging
import os
from collections import OrderedDict

from gemmi import cif

logger = logging.getLogger(__name__)


class StarFileError(Exception):
    pass


class StarFile:
    def __init__(self, filepath="", blocks=None):
        """
        Initialize either from a path to a STAR file or from an OrderedDict of ndarrays
        """
        # If blocks are given, set self.blocks, otherwise initialize an empty OrderedDict()
        # The following code repacks provided blocks as strings while handling both singleton and iterable inputs.
        if blocks is not None:
            for k in blocks:
                for _k, _v in blocks[k].items():
                    if not isinstance(_v, str):
                        try:
                            # Cast iterable elements to string.
                            blocks[k][_k] = list(map(str, iter(_v)))
                        except TypeError:
                            # Singleton, cast to string.
                            blocks[k][_k] = str(_v)

        self.blocks = blocks or OrderedDict()
        self.filepath = str(filepath)
        # raise an error if blocks and a filepath are both passed
        if bool(self.filepath) and bool(len(self.blocks)):
            raise StarFileError(
                "Pass either a path to a STAR file or an OrderedDict of dicts, not both"
            )
        if self.filepath:
            if not os.path.exists(self.filepath):
                logger.error(f"Could not open {self.filepath}")
                raise FileNotFoundError
            self._initialize_blocks()
        logger.info(f"Created {self}")

    def _initialize_blocks(self):
        """
        Converts a gemmi Document object representing the .star file
        at self.filepath into an OrderedDict of dicts, each of which represents one block in the .star file
        """
        logger.info(f"Parsing star file at: {self.filepath}")
        gemmi_doc = cif.read_file(self.filepath)
        # iterate over gemmi Block objects in the gemmi Document
        for gemmi_block in gemmi_doc:
            # iterating over gemmi Block objects yields Item objects
            # Items can have a Loop object and/or a Pair object
            # Loops correspond to the regular loop_ structure in a STAR file
            # Pairs have type List[str[2]] and correspond to a non-loop key value
            # pair in a STAR file, e.g.
            # _field1 \t 'value' #1

            # Our model of the .star file only allows a block to be one or the other
            block_has_pair = False
            block_has_loop = False
            # populated if this block has a pair
            pairs = {}
            # populated if this block as a loop
            loop_tags = []
            loop_data = []
            # Correct for GEMMI default behavior.
            # If a block is called 'data_' in the .star file:
            #   gemmi>=0.6.2 names it ' '
            #   gemmi<0.6.2 names it '#'
            # Rename it '' consistently.
            if gemmi_block.name in (" ", "#"):
                gemmi_block.name = ""
            for gemmi_item in gemmi_block:
                if gemmi_item.pair is not None:
                    block_has_pair = True
                    # if we find both a pair and a loop raise an error
                    if block_has_loop:
                        raise StarFileError(
                            "Blocks with multiple loops and/or pairs are not supported"
                        )
                    # assign key-value pair to dictionary
                    pair_key, pair_val = gemmi_item.pair
                    if pair_key not in pairs:
                        # read in as str because we do not want type conversion
                        pairs[pair_key] = str(pair_val)
                    else:
                        raise StarFileError(
                            f"Duplicate key in pair: {gemmi_item.pair[0]}"
                        )
                if gemmi_item.loop is not None:
                    block_has_loop = True
                    # if we find both a pair and a loop raise an error
                    if block_has_pair:
                        raise StarFileError(
                            "Blocks with multiple loops and/or pairs are not supported"
                        )
                    loop_tags = gemmi_item.loop.tags
                    # convert loop data to a list of lists
                    # using the .val(row, col) method of gemmi's Loop class
                    loop_data = [None] * gemmi_item.loop.length()
                    for row in range(gemmi_item.loop.length()):
                        loop_data[row] = [
                            gemmi_item.loop.val(row, col)
                            for col in range(gemmi_item.loop.width())
                        ]
            if block_has_pair:
                if gemmi_block.name not in self.blocks:
                    # represent a set of pairs by a dictionary
                    self.blocks[gemmi_block.name] = pairs
                else:
                    # enforce unique block names (keys of StarFile.block OrderedDict)
                    raise StarFileError(
                        f"Attempted overwrite of existing data block: {gemmi_block.name}"
                    )
            elif block_has_loop:
                if gemmi_block.name not in self.blocks:
                    # initialize dict from list of lists
                    d = {}
                    for i, loop_tag in enumerate(loop_tags):
                        d[loop_tag] = [str(x[i]) for x in loop_data]
                    self.blocks[gemmi_block.name] = d
                else:
                    # enforce unique block names (keys of StarFile.block OrderedDict)
                    raise StarFileError(
                        f"Attempted overwrite of existing data block: {gemmi_block.name}"
                    )

    def write(self, filepath):
        """
        Converts `blocks` to a gemmi Document and writes to a starfile at the given filepath.
        """
        # create an empty Document
        _doc = cif.Document()
        filepath = str(filepath)
        for name, block in self.blocks.items():
            # construct new empty block
            _block = _doc.add_new_block(name)
            # if this block (loop or pair) is empty, continue
            if len(block) == 0:
                continue

            # look at values to detect if we're dealing with iterables
            multiple_rows = isinstance(block, dict) and all(
                not isinstance(v, str) and hasattr(v, "__iter__")
                for v in block.values()
            )

            if not multiple_rows:
                for key, value in block.items():
                    # simply assign one pair item for each dict entry
                    # write out as str because we do not want type conversion
                    _block.set_pair(key, str(value))
            else:
                assert (
                    len(set(len(v) for v in block.values())) == 1
                ), "Not all iterables of the block have the same length"
                _n_rows = len(list(block.values())[0])
                # initialize loop with column names
                _keys = list(block.keys())
                _loop = _block.init_loop("", _keys)
                for i in range(_n_rows):
                    row = [str(block[k][i]) for k in _keys]
                    _loop.add_row(row)

        _doc.write_file(filepath)

    def get_block_by_index(self, index):
        """
        Retrieve a dict representing a star file block by its position in the starfile
        """
        return self.blocks[list(self.blocks.keys())[index]]

    def __getitem__(self, key):
        """
        Retrieve the star file block with name `key`
        """
        return self.blocks[key]

    def __setitem__(self, key, value):
        """
        Pass in a Pandas Dataframe or dictionary to add a block named `key` with values `value`
        """
        self.blocks[key] = value

    def __repr__(self):
        return "StarFile with blocks: " + ", ".join(self.blocks.keys())

    def __iter__(self):
        return self.blocks.items().__iter__()

    def __len__(self):
        return len(self.blocks)

    def __eq__(self, other):
        if not len(self) == len(other):
            return False
        self_list = list(self.blocks.items())
        other_list = list(other.blocks.items())
        for i in range(len(self_list)):
            # test whether block names are the same
            if not self_list[i][0] == other_list[i][0]:
                return False
            # test whether blocks are both DFs or dicts
            if not isinstance(self_list[i][1], type(other_list[i][1])):
                return False
            # finally, compare the objects themselves
            if not self_list[i][1] == other_list[i][1]:
                return False
        return True



import logging
import os
from concurrent import futures
from glob import glob

import numpy as np
from PIL import Image

from aspire.apple.picking import Picker
from aspire.utils import tqdm

logger = logging.getLogger(__name__)


class Apple:
    def __init__(
        self,
        particle_size,
        output_dir="",
        min_particle_size=None,
        max_particle_size=None,
        query_image_size=None,
        minimum_overlap_amount=None,
        tau1=None,
        tau2=None,
        container_size=450,
        model="svm",
        model_opts=None,
        mrc_margin_left=99,
        mrc_margin_right=100,
        mrc_margin_top=99,
        mrc_margin_bottom=100,
        mrc_shrink_factor=2,
        mrc_gauss_filter_size=15,
        mrc_gauss_filter_sigma=0.5,
        response_thresh_norm_factor=20,
        conv_map_nthreads=4,
        n_processes=1,
    ):
        """
        Instantiate an Apple instance with a given configuration.

        Some discussion of parameters can be found in:

            APPLE picker : Automatic particle picking, a low-effort cryo-EM framework.
            Heimowitz, Ayelet; Andén, Joakim; Singer, Amit.
            Journal of Structural Biology, Vol. 204, No. 2, 11.2018, p. 215-227.

        :param particle_size: Particle size (pixels) is required.
            Remaining parameters generally have defaults based on particle size.
        :param min_particle_size:
        :param max_particle_size:
        :param query_image_size:
        :param minimum_overlap_amount:
        :param tau1: SVM parameter
        :param tau2: SVM parameter
        :param container_size: Defaults 450
        :param output_dir: Optionally specify output_dir, defaults to local dir.
        :param model: One of svm/gaussian_mixture/gaussian_naive_bayes/xgboost/thunder_svm
        :param model_opts: Optional dictionary of svm model options. Defaults to None.
        :param n_processes: Concurrent processes to spawn.
            May improve performance on very large machines.
            Otherwise use default of 1.
        """

        self.particle_size = particle_size

        self.max_particle_size = max_particle_size or self.particle_size * 2

        self.min_particle_size = min_particle_size or self.particle_size // 4

        if self.max_particle_size < self.min_particle_size:
            raise RuntimeError("max_particle_size must be >= min_particle_size")

        self.minimum_overlap_amount = minimum_overlap_amount or self.particle_size // 10

        self.container_size = container_size
        self.n_processes = int(n_processes)
        self.output_dir = output_dir

        if query_image_size is None:
            query_image_size = np.round(self.particle_size * 2 / 3)
            query_image_size -= query_image_size % 4
            query_image_size = int(query_image_size)

        self.query_image_size = query_image_size

        # q_box is the query box (or "container") size
        self.q_box = (4000**2) / (self.query_image_size**2) * 4
        # tau1 and tau2 correspond to particle and noise windows respectively
        self.tau1 = tau1 or int(self.q_box * 0.03)
        self.tau2 = tau2 or int(self.q_box * 0.3)

        if self.output_dir and not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        self.verify_input_values()

        self.model = model
        self.model_opts = model_opts

        # MRC processing config
        # Margins to discard from any processed .mrc file
        # TODO: Margins are asymmetrical to conform to old behavior - fix going forward
        self.mrc_margin_left = mrc_margin_left
        self.mrc_margin_right = mrc_margin_right
        self.mrc_margin_top = mrc_margin_top
        self.mrc_margin_bottom = mrc_margin_bottom
        self.mrc_shrink_factor = mrc_shrink_factor
        self.mrc_gauss_filter_size = mrc_gauss_filter_size
        self.mrc_gauss_filter_sigma = mrc_gauss_filter_sigma
        self.response_thresh_norm_factor = response_thresh_norm_factor
        self.conv_map_nthreads = conv_map_nthreads

        # Assigned when process_micrograph_centers is run
        self.picker = dict()

    def verify_input_values(self):
        assert (
            1 <= self.max_particle_size <= 3000
        ), "Max particle size must be in range [1, 3000]!"

        assert (
            1 <= self.query_image_size <= 3000
        ), "Query image size must be in range [1, 3000]!"

        assert (
            5 <= self.particle_size < 3000
        ), "Particle size must be in range [5, 3000]!"

        assert (
            1 <= self.min_particle_size < 3000
        ), "Min particle size must be in range [1, 3000]!"

        max_tau1_value = self.q_box
        assert (
            0 <= self.tau1 <= max_tau1_value
        ), f"tau1 must be a in range [0, {max_tau1_value}]!"

        max_tau2_value = self.q_box
        assert (
            0 <= self.tau2 <= max_tau2_value
        ), f"tau2 must be in range [0, {max_tau2_value}]!"

        assert (
            0 <= self.minimum_overlap_amount <= 3000
        ), "overlap must be in range [0, 3000]!"

        # max container_size condition is (container_size_max * 2 + 200 > 4000), which is 1900
        assert (
            self.particle_size <= self.container_size <= 1900
        ), f"Container size must be within range [{self.particle_size}, 1900]!"

        assert (
            self.particle_size >= self.query_image_size
        ), f"Particle size ({self.particle_size}) must exceed query image size ({self.query_image_size})!"

    def process_folder(self, folder, create_jpg=False):
        # Gather matches for multiple possible file types.
        filenames = []
        for ext in Image.extensions:
            filenames.extend(glob(f"{folder}/*{ext}"))
        logger.info(f"converting {len(filenames)} input files")
        logger.info(f"launching {self.n_processes} processes")

        pbar = tqdm(total=len(filenames))
        with futures.ProcessPoolExecutor(self.n_processes) as executor:
            to_do = []
            for filename in filenames:
                future = executor.submit(
                    self.process_micrograph, filename, False, create_jpg
                )
                to_do.append(future)

            for future in futures.as_completed(to_do):
                # Retrieve (None) result, since this operation re-raises Exceptions, if any.
                _ = future.result()
                pbar.update(1)
        pbar.close()

    def process_micrograph_centers(
        self,
        filepath,
    ):
        """
        Process micrograph at `filepath`, returning `centers`.

        :param filepath: MRC filepath
        :return: `centers`
        """

        # Note we assign picker now so we can access it from process_micrograph_plots
        self.picker[filepath] = picker = Picker(
            self.particle_size,
            self.max_particle_size,
            self.min_particle_size,
            self.query_image_size,
            self.tau1,
            self.tau2,
            self.minimum_overlap_amount,
            self.container_size,
            filepath,
            self.output_dir,
            model=self.model,
            model_opts=self.model_opts,
            mrc_margin_left=self.mrc_margin_left,
            mrc_margin_right=self.mrc_margin_right,
            mrc_margin_top=self.mrc_margin_top,
            mrc_margin_bottom=self.mrc_margin_bottom,
            mrc_shrink_factor=self.mrc_shrink_factor,
            mrc_gauss_filter_size=self.mrc_gauss_filter_size,
            mrc_gauss_filter_sigma=self.mrc_gauss_filter_sigma,
            response_thresh_norm_factor=self.response_thresh_norm_factor,
            conv_map_nthreads=self.conv_map_nthreads,
        )

        logger.info("Computing scores for query images")
        score = (
            picker.query_score()
        )  # compute score using normalized cross-correlations

        while True:
            logger.info(f"Running svm with tau1={picker.tau1}, tau2={picker.tau2}")
            # train SVM classifier and classify all windows in micrograph
            segmentation = picker.run_svm(score)

            # If all windows are classified identically, update tau_1 or tau_2
            if np.all(segmentation):
                picker.tau2 += 500
            elif not np.any(segmentation):
                picker.tau1 += 500
            else:
                break

        logger.info("Discarding suspected artifacts")
        segmentation = picker.morphology_ops(segmentation)

        logger.info("Getting particle centers")
        centers = picker.extract_particles(segmentation)

        return centers

    def process_micrograph_plots(self, filepath, centers, create_jpg=False):
        """
        Takes in `centers`, returns corresponding `particle_image`.

        Optionally writes jpg to disk.

        :param filepath: mrc filepath
        :param centers: Particle centers, typically from `process_micrograph_centers`.
        :param create_jpg: Optionally writes JPG file with picked particles.
        :return: `particle_image`
        """

        # Note that we only use filepath to identify the specific picker,
        #   since Apple instance may be shared.
        picker = self.picker.get(filepath)
        if picker is None:
            raise RuntimeError(
                "Must run `process_micrograph_centers` for this filepath first"
            )

        particle_image = self.particle_image(
            picker.original_im, picker.particle_size, centers
        )

        output_dir = self.output_dir

        # If user wants to create_jpg, but has None output_dir,
        #  warn them and use local dir instead of crashing.
        if create_jpg and output_dir is None:
            output_dir = "."
            logger.warn(
                "`create_jpg` called with (default) output_dir=None."
                f'  Using output_dir = "{output_dir}".'
                "  To avoid this warning, set `output_dir` when creating JPG files."
            )

        if create_jpg:
            # Create the image
            image_out = Image.fromarray((particle_image).astype(np.uint8))

            # Construct filename
            base = os.path.splitext(os.path.basename(picker.filename))[0]
            filename_out = os.path.join(output_dir, f"{base}_result.jpg")

            # Save the image
            image_out.save(filename_out)

        return particle_image

    def process_micrograph(self, filepath, create_jpg=False):
        """
        Process micrograph at `filepath`, returning (`centers`, `particle_image`).

        :param filepath: mrc filepath
        :param create_jpg: Optionally writes JPG file identifying picked particles.
        :return: (`centers`, `particle_image`)
        """

        centers = self.process_micrograph_centers(filepath)
        particle_image = self.process_micrograph_plots(filepath, centers, create_jpg)
        return centers, particle_image

    def particle_image(self, micro_img, particle_size, centers):
        """
        Return a numpy array representing the picked centers on a micrograph,
        suitable for rendering in a jupyter notebook or saving as a jpg etc.

        :param micro_img: The micrograph image as a numpy array
        :param particle_size: Particle size of picked particles
        :param centers: Picked centers for micrograph.
        :return: A numpy array with picked centers displayed as rectangles
        """

        micro_img = micro_img - np.amin(
            np.reshape(micro_img, (np.prod(micro_img.shape)))
        )
        picks = np.ones(micro_img.shape)
        for i in range(0, centers.shape[0]):
            y = int(centers[i, 1])
            x = int(centers[i, 0])
            d = int(np.floor(particle_size))
            picks[y - d : y - d + 5, x - d : x + d] = 0
            picks[y + d : y + d + 5, x - d : x + d] = 0
            picks[y - d : y + d, x - d : x - d + 5] = 0
            picks[y - d : y + d, x + d : x + d + 5] = 0

        return np.multiply(micro_img, picks)
